from __future__ import annotations

import asyncio
import sys
import warnings
from contextlib import suppress
from datetime import datetime, timedelta, timezone
from functools import cached_property
from typing import TYPE_CHECKING, Any, Literal, TypeVar, cast, overload

from lazy_object_proxy import Proxy
from more_itertools import flatten
from pydantic import AliasChoices

from apify_client import ApifyClientAsync
from apify_shared.consts import ActorEnvVars, ActorExitCodes, ApifyEnvVars
from crawlee import service_locator
from crawlee.errors import ServiceConflictError
from crawlee.events import (
    Event,
    EventAbortingData,
    EventExitData,
    EventListener,
    EventMigratingData,
    EventPersistStateData,
    EventSystemInfoData,
)

from apify._charging import DEFAULT_DATASET_ITEM_EVENT, ChargeResult, ChargingManager, ChargingManagerImplementation
from apify._configuration import Configuration
from apify._consts import EVENT_LISTENERS_TIMEOUT
from apify._crypto import decrypt_input_secrets, load_private_key
from apify._models import ActorRun
from apify._proxy_configuration import ProxyConfiguration
from apify._utils import docs_group, docs_name, ensure_context, get_system_info, is_running_in_ipython
from apify.events import ApifyEventManager, EventManager, LocalEventManager
from apify.log import _configure_logging, logger
from apify.storage_clients import ApifyStorageClient, SmartApifyStorageClient
from apify.storage_clients._file_system import ApifyFileSystemStorageClient
from apify.storages import Dataset, KeyValueStore, RequestQueue

if TYPE_CHECKING:
    import logging
    from collections.abc import Callable
    from types import TracebackType

    from typing_extensions import Self

    from crawlee._types import JsonSerializable
    from crawlee.proxy_configuration import _NewUrlFunction

    from apify._models import Webhook

MainReturnType = TypeVar('MainReturnType')

_ensure_context = ensure_context('_active')


@docs_name('Actor')
@docs_group('Actor')
class _ActorType:
    """The core class for building Actors on the Apify platform.

    Actors are serverless programs running in the cloud that can perform anything from simple actions
    (such as filling out a web form or sending an email) to complex operations (such as crawling an
    entire website or removing duplicates from a large dataset). They are packaged as Docker containers
    which accept well-defined JSON input, perform an action, and optionally produce well-defined output.

    ### References

    - Apify platform documentation: https://docs.apify.com/platform/actors
    - Actor whitepaper: https://whitepaper.actor/

    ### Usage

    ```python
    import asyncio

    import httpx
    from apify import Actor
    from bs4 import BeautifulSoup


    async def main() -> None:
        async with Actor:
            actor_input = await Actor.get_input()
            async with httpx.AsyncClient() as client:
                response = await client.get(actor_input['url'])
            soup = BeautifulSoup(response.content, 'html.parser')
            data = {
                'url': actor_input['url'],
                'title': soup.title.string if soup.title else None,
            }
            await Actor.push_data(data)

    if __name__ == '__main__':
        asyncio.run(main())
    ```
    """

    _ACTOR_STATE_KEY = 'APIFY_GLOBAL_STATE'

    def __init__(
        self,
        configuration: Configuration | None = None,
        *,
        configure_logging: bool = True,
        exit_process: bool | None = None,
        exit_code: int = 0,
        status_message: str | None = None,
        event_listeners_timeout: timedelta | None = EVENT_LISTENERS_TIMEOUT,
        cleanup_timeout: timedelta = timedelta(seconds=30),
    ) -> None:
        """Initialize a new instance.

        Args:
            configuration: The Actor configuration to use. If not provided, a default configuration is created.
            configure_logging: Whether to set up the default logging configuration.
            exit_process: Whether the Actor should call `sys.exit` when the context manager exits.
                Defaults to True, except in IPython, Pytest, and Scrapy environments.
            exit_code: The exit code the Actor should use when exiting.
            status_message: Final status message to display upon Actor termination.
            event_listeners_timeout: Maximum time to wait for Actor event listeners to complete before exiting.
            cleanup_timeout: Maximum time to wait for cleanup tasks to finish.
        """
        self._configuration = configuration
        self._configure_logging = configure_logging
        self._exit_process = self._get_default_exit_process() if exit_process is None else exit_process
        self._exit_code = exit_code
        self._status_message = status_message
        self._event_listeners_timeout = event_listeners_timeout
        self._cleanup_timeout = cleanup_timeout

        # Actor state when this method is being executed is unpredictable.
        # Actor can be initialized by lazy object proxy or by user directly, or by both.
        # Until `init` method is run, this state of uncertainty remains. This is the reason why any setting done here in
        # `__init__` method should not be considered final.

        self._apify_client: ApifyClientAsync | None = None

        # Keep track of all used state stores to persist their values on exit
        self._use_state_stores: set[str | None] = set()

        self._active = False
        """Whether the Actor instance is currently active (initialized and within context)."""

        self._is_rebooting = False
        """Whether the Actor is currently rebooting."""

        self._is_exiting = False
        """Whether the Actor is currently exiting."""

    async def __aenter__(self) -> Self:
        """Enter the Actor context.

        Initializes the Actor when used in an `async with` block. This method:

        - Sets up local or cloud storage clients depending on whether the Actor runs locally or on the Apify platform.
        - Configures the event manager and starts periodic state persistence.
        - Initializes the charging manager for handling charging events.
        - Configures logging after all core services are registered.

        This method must be called exactly once per Actor instance. Re-initializing an Actor or having multiple
        active Actor instances is not standard usage and may lead to warnings or unexpected behavior.
        """
        if self._active:
            raise RuntimeError('The Actor was already initialized!')

        # Initialize configuration first - it's required for the next steps.
        if self._configuration:
            # User provided explicit configuration - register it in the service locator.
            service_locator.set_configuration(self.configuration)
        else:
            # No explicit configuration provided - trigger creation of default configuration.
            _ = self.configuration

        # Configure logging based on the configuration, any logs before this point are lost.
        if self._configure_logging:
            _configure_logging()
            self.log.debug('Logging configured')

        self.log.info('Initializing Actor', extra=get_system_info())
        self.log.debug('Configuration initialized')

        # Update the global Actor proxy to refer to this instance.
        cast('Proxy', Actor).__wrapped__ = self  # ty: ignore[invalid-assignment]
        self._is_exiting = False
        self._was_final_persist_state_emitted = False

        # Initialize the storage client and register it in the service locator.
        _ = self._storage_client
        self.log.debug('Storage client initialized')

        # Initialize the event manager and register it in the service locator.
        await self.event_manager.__aenter__()
        self.log.debug('Event manager initialized')

        # Initialize the charging manager.
        await self._charging_manager_implementation.__aenter__()
        self.log.debug('Charging manager initialized')

        # Mark initialization as complete and update global state.
        self._active = True

        if not Actor.is_at_home():
            # Make sure that the input related KVS is initialized to ensure that the input aware client is used
            await self.open_key_value_store()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
    ) -> None:
        """Exit the Actor context.

        If the block exits with an exception, the Actor fails with a non-zero exit code.
        Otherwise, it exits cleanly. In both cases the Actor:

        - Cancels periodic `PERSIST_STATE` events.
        - Sends a final `PERSIST_STATE` event.
        - Waits for all event listeners to finish.
        - Stops the event manager and the charging manager.
        - Optionally terminates the process with the selected exit code.
        """
        if self._is_exiting:
            return

        if not self._active:
            raise RuntimeError('The _ActorType is not active. Use it within the async context.')

        if exc_value and not is_running_in_ipython():
            # In IPython, we don't run `sys.exit()` during Actor exits,
            # so the exception traceback will be printed on its own
            self.log.exception('Actor failed with an exception', exc_info=exc_value)
            self.exit_code = ActorExitCodes.ERROR_USER_FUNCTION_THREW.value

        self._is_exiting = True
        self.log.info('Exiting Actor', extra={'exit_code': self.exit_code})

        async def finalize() -> None:
            if self.status_message is not None:
                await self.set_status_message(self.status_message, is_terminal=True)

            # Sleep for a bit so that the listeners have a chance to trigger
            await asyncio.sleep(0.1)

            if self._event_listeners_timeout:
                await self.event_manager.wait_for_all_listeners_to_complete(timeout=self._event_listeners_timeout)

            await self.event_manager.__aexit__(None, None, None)
            await self._charging_manager_implementation.__aexit__(None, None, None)

            # Persist Actor state
            await self._save_actor_state()

        try:
            await asyncio.wait_for(finalize(), self._cleanup_timeout.total_seconds())
        except TimeoutError:
            self.log.exception('Actor cleanup timed out')
        finally:
            self._active = False

        if self._exit_process:
            sys.exit(self.exit_code)

    def __repr__(self) -> str:
        if self is cast('Proxy', Actor).__wrapped__:
            return '<apify.Actor>'

        return super().__repr__()

    def __call__(
        self,
        configuration: Configuration | None = None,
        *,
        configure_logging: bool = True,
        exit_process: bool | None = None,
        exit_code: int = 0,
        event_listeners_timeout: timedelta | None = EVENT_LISTENERS_TIMEOUT,
        status_message: str | None = None,
        cleanup_timeout: timedelta = timedelta(seconds=30),
    ) -> Self:
        """Make a new Actor instance with a non-default configuration.

        This is necessary due to the lazy object proxying of the global `Actor` instance.
        """
        return self.__class__(
            configuration=configuration,
            configure_logging=configure_logging,
            exit_process=exit_process,
            exit_code=exit_code,
            event_listeners_timeout=event_listeners_timeout,
            status_message=status_message,
            cleanup_timeout=cleanup_timeout,
        )

    @property
    def log(self) -> logging.Logger:
        """Logger configured for this Actor."""
        return logger

    @property
    def exit_code(self) -> int:
        """The exit code the Actor will use when exiting."""
        return self._exit_code

    @exit_code.setter
    def exit_code(self, value: int) -> None:
        self._exit_code = value

    @property
    def status_message(self) -> str | None:
        """The final status message that the Actor will display upon termination."""
        return self._status_message

    @status_message.setter
    def status_message(self, value: str | None) -> None:
        self._status_message = value

    @property
    def apify_client(self) -> ApifyClientAsync:
        """Asynchronous Apify client for interacting with the Apify API."""
        if not self._apify_client:
            self._apify_client = self.new_client()
        return self._apify_client

    @cached_property
    def configuration(self) -> Configuration:
        """Actor configuration, uses the default instance if not explicitly set."""
        if self._configuration:
            return self._configuration

        try:
            # Set implicit default Apify configuration, unless configuration was already set.
            implicit_configuration = Configuration()
            service_locator.set_configuration(implicit_configuration)
            self._configuration = implicit_configuration
        except ServiceConflictError:
            self.log.debug(
                'Configuration in service locator was set explicitly before Actor.init was called.'
                'Using the existing configuration as implicit configuration for the Actor.'
            )

        # Use the configuration from the service locator
        self._configuration = Configuration.get_global_configuration()
        return self._configuration

    @cached_property
    def event_manager(self) -> EventManager:
        """Manages Apify platform events.

        It uses `ApifyEventManager` on the Apify platform and `LocalEventManager` otherwise.
        """
        try:
            event_manager = (
                ApifyEventManager(
                    configuration=self.configuration,
                    persist_state_interval=self.configuration.persist_state_interval,
                )
                if self.is_at_home()
                else LocalEventManager(
                    system_info_interval=self.configuration.system_info_interval,
                    persist_state_interval=self.configuration.persist_state_interval,
                )
            )
            service_locator.set_event_manager(event_manager)
        except ServiceConflictError:
            self.log.debug(
                'Event manager already exists in service locator (set by previous Actor context or explicitly by '
                'user). Using the existing event manager.'
            )
            # Use the event manager from the service locator
            event_manager = service_locator.get_event_manager()

        return event_manager

    @cached_property
    def _charging_manager_implementation(self) -> ChargingManagerImplementation:
        return ChargingManagerImplementation(self.configuration, self.apify_client)

    @cached_property
    def _storage_client(self) -> SmartApifyStorageClient:
        """Storage client used by the Actor.

        Depending on the initialization of the service locator the client can be created in different ways.
        """
        try:
            # Nothing was set by the user.
            implicit_storage_client = SmartApifyStorageClient(
                local_storage_client=ApifyFileSystemStorageClient(), cloud_storage_client=ApifyStorageClient()
            )
            service_locator.set_storage_client(implicit_storage_client)
        except ServiceConflictError:
            self.log.debug(
                'Storage client in service locator was set explicitly before Actor.init was called. '
                'Using the existing storage client as implicit storage client for the Actor.'
            )
        else:
            return implicit_storage_client

        # User set something in the service locator.
        explicit_storage_client = service_locator.get_storage_client()
        if isinstance(explicit_storage_client, SmartApifyStorageClient):
            # The client was manually set to the right type in the service locator. This is the explicit way.
            return explicit_storage_client

        raise RuntimeError(
            'The storage client in the service locator has to be instance of SmartApifyStorageClient. If you want to '
            'set the storage client manually you have to call '
            '`service_locator.set_storage_client(SmartApifyStorageClient(...))` before entering Actor context or '
            'awaiting `Actor.init`.'
        )

    async def init(self) -> None:
        """Initialize the Actor without using context-manager syntax.

        Equivalent to `await Actor.__aenter__()`.
        """
        await self.__aenter__()

    async def exit(
        self,
        *,
        exit_code: int = 0,
        status_message: str | None = None,
        event_listeners_timeout: timedelta | None = EVENT_LISTENERS_TIMEOUT,
        cleanup_timeout: timedelta = timedelta(seconds=30),
    ) -> None:
        """Exit the Actor without using context-manager syntax.

        Equivalent to `await Actor.__aexit__()`.

        Args:
            exit_code: The exit code the Actor should use when exiting.
            status_message: Final status message to display upon Actor termination.
            event_listeners_timeout: Maximum time to wait for Actor event listeners to complete before exiting.
            cleanup_timeout: Maximum time to wait for cleanup tasks to finish.
        """
        self.exit_code = exit_code
        self.status_message = status_message
        self._event_listeners_timeout = event_listeners_timeout
        self._cleanup_timeout = cleanup_timeout
        await self.__aexit__(None, None, None)

    async def fail(
        self,
        *,
        exit_code: int = 1,
        exception: BaseException | None = None,
        status_message: str | None = None,
    ) -> None:
        """Fail the Actor instance without using context-manager syntax.

        Equivalent to setting the `self.exit_code` and `self.status_message` properties and using
        `await Actor.__aexit__()`.

        Args:
            exit_code: The exit code with which the Actor should fail (defaults to `1`).
            exception: The exception with which the Actor failed.
            status_message: The final status message that the Actor should display.
        """
        self.exit_code = exit_code
        self.status_message = status_message

        await self.__aexit__(
            exc_type=type(exception) if exception else None,
            exc_value=exception,
            exc_traceback=exception.__traceback__ if exception else None,
        )

    def new_client(
        self,
        *,
        token: str | None = None,
        api_url: str | None = None,
        max_retries: int | None = None,
        min_delay_between_retries: timedelta | None = None,
        timeout: timedelta | None = None,
    ) -> ApifyClientAsync:
        """Return a new instance of the Apify API client.

        The `ApifyClientAsync` class is provided by the [apify-client](https://github.com/apify/apify-client-python)
        package, and it is automatically configured using the `APIFY_API_BASE_URL` and `APIFY_TOKEN` environment
        variables.

        You can override the token via the available options. That's useful if you want to use the client
        as a different Apify user than the SDK internals are using.

        Args:
            token: The Apify API token.
            api_url: The URL of the Apify API server to which to connect to. Defaults to https://api.apify.com.
            max_retries: How many times to retry a failed request at most.
            min_delay_between_retries: How long will the client wait between retrying requests
                (increases exponentially from this value).
            timeout: The socket timeout of the HTTP requests sent to the Apify API.
        """
        token = token or self.configuration.token
        api_url = api_url or self.configuration.api_base_url
        return ApifyClientAsync(
            token=token,
            api_url=api_url,
            max_retries=max_retries,
            min_delay_between_retries_millis=int(min_delay_between_retries.total_seconds() * 1000)
            if min_delay_between_retries is not None
            else None,
            timeout_secs=int(timeout.total_seconds()) if timeout else None,
        )

    @_ensure_context
    async def open_dataset(
        self,
        *,
        id: str | None = None,
        alias: str | None = None,
        name: str | None = None,
        force_cloud: bool = False,
    ) -> Dataset:
        """Open a dataset.

        Datasets are used to store structured data where each object stored has the same attributes, such as online
        store products or real estate offers. The actual data is stored either on the local filesystem or in
        the Apify cloud.

        Args:
            id: The ID of the dataset to open. If provided, searches for existing dataset by ID.
                Mutually exclusive with name and alias.
            name: The name of the dataset to open (global scope, persists across runs).
                Mutually exclusive with id and alias.
            alias: The alias of the dataset to open (run scope, creates unnamed storage).
                Mutually exclusive with id and name.
            force_cloud: If set to `True` then the Apify cloud storage is always used. This way it is possible
                to combine local and cloud storage.

        Returns:
            An instance of the `Dataset` class for the given ID or name.
        """
        return await Dataset.open(
            id=id,
            name=name,
            alias=alias,
            storage_client=self._storage_client.get_suitable_storage_client(force_cloud=force_cloud),
        )

    @_ensure_context
    async def open_key_value_store(
        self,
        *,
        id: str | None = None,
        alias: str | None = None,
        name: str | None = None,
        force_cloud: bool = False,
    ) -> KeyValueStore:
        """Open a key-value store.

        Key-value stores are used to store records or files, along with their MIME content type. The records are stored
        and retrieved using a unique key. The actual data is stored either on a local filesystem or in the Apify cloud.

        Args:
            id: The ID of the KVS to open. If provided, searches for existing KVS by ID.
                Mutually exclusive with name and alias.
            name: The name of the KVS to open (global scope, persists across runs).
                Mutually exclusive with id and alias.
            alias: The alias of the KVS to open (run scope, creates unnamed storage).
                Mutually exclusive with id and name.
            force_cloud: If set to `True` then the Apify cloud storage is always used. This way it is possible
                to combine local and cloud storage.

        Returns:
            An instance of the `KeyValueStore` class for the given ID or name.
        """
        return await KeyValueStore.open(
            id=id,
            name=name,
            alias=alias,
            storage_client=self._storage_client.get_suitable_storage_client(force_cloud=force_cloud),
        )

    @_ensure_context
    async def open_request_queue(
        self,
        *,
        id: str | None = None,
        alias: str | None = None,
        name: str | None = None,
        force_cloud: bool = False,
    ) -> RequestQueue:
        """Open a request queue.

        Request queue represents a queue of URLs to crawl, which is stored either on local filesystem or in
        the Apify cloud. The queue is used for deep crawling of websites, where you start with several URLs and then
        recursively follow links to other pages. The data structure supports both breadth-first and depth-first
        crawling orders.

        Args:
            id: The ID of the RQ to open. If provided, searches for existing RQ by ID.
                Mutually exclusive with name and alias.
            name: The name of the RQ to open (global scope, persists across runs).
                Mutually exclusive with id and alias.
            alias: The alias of the RQ to open (run scope, creates unnamed storage).
                Mutually exclusive with id and name.
            force_cloud: If set to `True` then the Apify cloud storage is always used. This way it is possible
                to combine local and cloud storage.

        Returns:
            An instance of the `RequestQueue` class for the given ID or name.
        """
        return await RequestQueue.open(
            id=id,
            name=name,
            alias=alias,
            storage_client=self._storage_client.get_suitable_storage_client(force_cloud=force_cloud),
        )

    @_ensure_context
    async def push_data(self, data: dict | list[dict], charged_event_name: str | None = None) -> ChargeResult:
        """Store an object or a list of objects to the default dataset of the current Actor run.

        Args:
            data: The data to push to the default dataset.
            charged_event_name: If provided and if the Actor uses the pay-per-event pricing model,
                the method will attempt to charge for the event for each pushed item.
        """
        if charged_event_name and charged_event_name.startswith('apify-'):
            raise ValueError(f'Cannot charge for synthetic event "{charged_event_name}" manually')

        charging_manager = self.get_charging_manager()

        if not data:
            charged_event_name = charged_event_name or DEFAULT_DATASET_ITEM_EVENT
            charge_limit_reached = charging_manager.is_event_charge_limit_reached(charged_event_name)

            return ChargeResult(
                event_charge_limit_reached=charge_limit_reached,
                charged_count=0,
                chargeable_within_limit=charging_manager.compute_chargeable(),
            )

        data = data if isinstance(data, list) else [data]

        dataset = await self.open_dataset()

        # Acquire the charge lock to prevent race conditions between concurrent
        # push_data calls. We need to hold the lock for the entire push_data + charge sequence.
        async with charging_manager.charge_lock():
            # Synthetic events are handled within dataset.push_data, only get data for `ChargeResult`.
            if charged_event_name is None:
                before = charging_manager.get_charged_event_count(DEFAULT_DATASET_ITEM_EVENT)
                await dataset.push_data(data)
                after = charging_manager.get_charged_event_count(DEFAULT_DATASET_ITEM_EVENT)
                return ChargeResult(
                    event_charge_limit_reached=charging_manager.is_event_charge_limit_reached(
                        DEFAULT_DATASET_ITEM_EVENT
                    ),
                    charged_count=after - before,
                    chargeable_within_limit=charging_manager.compute_chargeable(),
                )

            pushed_items_count = charging_manager.compute_push_data_limit(
                items_count=len(data),
                event_name=charged_event_name,
                is_default_dataset=True,
            )

            if pushed_items_count < len(data):
                await dataset.push_data(data[:pushed_items_count])
            elif pushed_items_count > 0:
                await dataset.push_data(data)

            # Only charge explicit events; synthetic events will be processed within the client.
            return await charging_manager.charge(
                event_name=charged_event_name,
                count=pushed_items_count,
            )

    @_ensure_context
    async def get_input(self) -> Any:
        """Get the Actor input value from the default key-value store associated with the current Actor run."""
        input_value = await self.get_value(self.configuration.input_key)
        input_secrets_private_key = self.configuration.input_secrets_private_key_file
        input_secrets_key_passphrase = self.configuration.input_secrets_private_key_passphrase
        if input_secrets_private_key and input_secrets_key_passphrase:
            private_key = load_private_key(
                input_secrets_private_key,
                input_secrets_key_passphrase,
            )
            input_value = decrypt_input_secrets(private_key, input_value)

        return input_value

    @_ensure_context
    async def get_value(self, key: str, default_value: Any = None) -> Any:
        """Get a value from the default key-value store associated with the current Actor run.

        Args:
            key: The key of the record which to retrieve.
            default_value: Default value returned in case the record does not exist.
        """
        key_value_store = await self.open_key_value_store()
        return await key_value_store.get_value(key, default_value)

    @_ensure_context
    async def set_value(
        self,
        key: str,
        value: Any,
        *,
        content_type: str | None = None,
    ) -> None:
        """Set or delete a value in the default key-value store associated with the current Actor run.

        Args:
            key: The key of the record which to set.
            value: The value of the record which to set, or None, if the record should be deleted.
            content_type: The content type which should be set to the value.
        """
        key_value_store = await self.open_key_value_store()
        return await key_value_store.set_value(key, value, content_type=content_type)

    @_ensure_context
    def get_charging_manager(self) -> ChargingManager:
        """Retrieve the charging manager to access granular pricing information."""
        return self._charging_manager_implementation

    @_ensure_context
    async def charge(self, event_name: str, count: int = 1) -> ChargeResult:
        """Charge for a specified number of events - sub-operations of the Actor.

        This is relevant only for the pay-per-event pricing model.

        Args:
            event_name: Name of the event to be charged for.
            count: Number of events to charge for.
        """
        # charging_manager.charge() acquires charge_lock internally.
        charging_manager = self.get_charging_manager()
        return await charging_manager.charge(event_name, count)

    @overload
    def on(
        self, event_name: Literal[Event.PERSIST_STATE], listener: EventListener[EventPersistStateData]
    ) -> EventListener[EventPersistStateData]: ...
    @overload
    def on(
        self, event_name: Literal[Event.SYSTEM_INFO], listener: EventListener[EventSystemInfoData]
    ) -> EventListener[EventSystemInfoData]: ...
    @overload
    def on(
        self, event_name: Literal[Event.MIGRATING], listener: EventListener[EventMigratingData]
    ) -> EventListener[EventMigratingData]: ...
    @overload
    def on(
        self, event_name: Literal[Event.ABORTING], listener: EventListener[EventAbortingData]
    ) -> EventListener[EventAbortingData]: ...
    @overload
    def on(
        self, event_name: Literal[Event.EXIT], listener: EventListener[EventExitData]
    ) -> EventListener[EventExitData]: ...
    @overload
    def on(self, event_name: Event, listener: EventListener[None]) -> EventListener[Any]: ...

    @_ensure_context
    def on(self, event_name: Event, listener: EventListener[Any]) -> EventListener[Any]:
        """Add an event listener to the Actor's event manager.

        The following events can be emitted:

        - `Event.SYSTEM_INFO`: Emitted every minute; the event data contains information about the Actor's resource
          usage.

        - `Event.MIGRATING`: Emitted when the Actor on the Apify platform is about to be migrated to another worker
          server. Use this event to persist the Actor's state and gracefully stop in-progress tasks, preventing
          disruption.

        - `Event.PERSIST_STATE`: Emitted regularly (default: 60 seconds) to notify the Actor to persist its state,
          preventing work repetition after a restart. This event is emitted together with the `MIGRATING` event, where
          the `isMigrating` flag in the event data is `True`; otherwise, the flag is `False`. This event is for
          convenience; the same effect can be achieved by setting an interval and listening for the `MIGRATING` event.

        - `Event.ABORTING`: Emitted when a user aborts an Actor run on the Apify platform, allowing the Actor time
          to clean up its state if the abort is graceful.

        Args:
            event_name: The Actor event to listen for.
            listener: The function to be called when the event is emitted (can be async).
        """
        self.event_manager.on(event=event_name, listener=listener)
        return listener

    @overload
    def off(self, event_name: Literal[Event.PERSIST_STATE], listener: EventListener[EventPersistStateData]) -> None: ...
    @overload
    def off(self, event_name: Literal[Event.SYSTEM_INFO], listener: EventListener[EventSystemInfoData]) -> None: ...
    @overload
    def off(self, event_name: Literal[Event.MIGRATING], listener: EventListener[EventMigratingData]) -> None: ...
    @overload
    def off(self, event_name: Literal[Event.ABORTING], listener: EventListener[EventAbortingData]) -> None: ...
    @overload
    def off(self, event_name: Literal[Event.EXIT], listener: EventListener[EventExitData]) -> None: ...
    @overload
    def off(self, event_name: Event, listener: EventListener[None]) -> None: ...

    @_ensure_context
    def off(self, event_name: Event, listener: Callable | None = None) -> None:
        """Remove a listener, or all listeners, from an Actor event.

        Args:
            event_name: The Actor event for which to remove listeners.
            listener: The listener which is supposed to be removed. If not passed, all listeners of this event
                are removed.
        """
        self.event_manager.off(event=event_name, listener=listener)

    def is_at_home(self) -> bool:
        """Return `True` when the Actor is running on the Apify platform, and `False` otherwise (e.g. local run)."""
        return self.configuration.is_at_home

    @_ensure_context
    def get_env(self) -> dict:
        """Return a dictionary with information parsed from all the `APIFY_XXX` environment variables.

        For a list of all the environment variables, see the
        [Actor documentation](https://docs.apify.com/actors/development/environment-variables). If some variables
        are not defined or are invalid, the corresponding value in the resulting dictionary will be None.
        """
        config = dict[str, Any]()
        for field_name, field in Configuration.model_fields.items():
            if field.deprecated:
                continue

            if field.alias:
                aliases = [field.alias]
            elif isinstance(field.validation_alias, str):
                aliases = [field.validation_alias]
            elif isinstance(field.validation_alias, AliasChoices):
                aliases = cast('list[str]', field.validation_alias.choices)
            else:
                aliases = [field_name]

            for alias in aliases:
                config[alias] = getattr(self.configuration, field_name)

        env_vars = {env_var.value.lower(): env_var.name.lower() for env_var in [*ActorEnvVars, *ApifyEnvVars]}
        return {option_name: config[env_var] for env_var, option_name in env_vars.items() if env_var in config}

    @_ensure_context
    async def start(
        self,
        actor_id: str,
        run_input: Any = None,
        *,
        token: str | None = None,
        content_type: str | None = None,
        build: str | None = None,
        memory_mbytes: int | None = None,
        timeout: timedelta | None | Literal['inherit', 'RemainingTime'] = None,
        wait_for_finish: int | None = None,
        webhooks: list[Webhook] | None = None,
    ) -> ActorRun:
        """Run an Actor on the Apify platform.

        Unlike `Actor.call`, this method just starts the run without waiting for finish.

        Args:
            actor_id: The ID of the Actor to be run.
            run_input: The input to pass to the Actor run.
            token: The Apify API token to use for this request (defaults to the `APIFY_TOKEN` environment variable).
            content_type: The content type of the input.
            build: Specifies the Actor build to run. It can be either a build tag or build number. By default,
                the run uses the build specified in the default run configuration for the Actor (typically latest).
            memory_mbytes: Memory limit for the run, in megabytes. By default, the run uses a memory limit specified
                in the default run configuration for the Actor.
            timeout: Optional timeout for the run, in seconds. By default, the run uses timeout specified in
                the default run configuration for the Actor. Using `inherit` or `RemainingTime` will set timeout of the
                other Actor to the time remaining from this Actor timeout.
            wait_for_finish: The maximum number of seconds the server waits for the run to finish. By default,
                it is 0, the maximum value is 300.
            webhooks: Optional ad-hoc webhooks (https://docs.apify.com/webhooks/ad-hoc-webhooks) associated with
                the Actor run which can be used to receive a notification, e.g. when the Actor finished or failed.
                If you already have a webhook set up for the Actor or task, you do not have to add it again here.

        Returns:
            Info about the started Actor run
        """
        client = self.new_client(token=token) if token else self.apify_client

        if webhooks:
            serialized_webhooks = [
                hook.model_dump(by_alias=True, exclude_unset=True, exclude_defaults=True) for hook in webhooks
            ]
        else:
            serialized_webhooks = None

        if timeout in {'inherit', 'RemainingTime'}:
            if timeout == 'RemainingTime':
                warnings.warn(
                    '`RemainingTime` is deprecated and will be removed in version 4.0.0. Use `inherit` instead.',
                    DeprecationWarning,
                    stacklevel=2,
                )
            actor_start_timeout = self._get_remaining_time()
        elif timeout is None:
            actor_start_timeout = None
        elif isinstance(timeout, timedelta):
            actor_start_timeout = timeout
        else:
            raise ValueError(
                f'Invalid timeout {timeout!r}: expected `None`, `"inherit"`, `"RemainingTime"`, or a `timedelta`.'
            )

        api_result = await client.actor(actor_id).start(
            run_input=run_input,
            content_type=content_type,
            build=build,
            memory_mbytes=memory_mbytes,
            timeout_secs=int(actor_start_timeout.total_seconds()) if actor_start_timeout is not None else None,
            wait_for_finish=wait_for_finish,
            webhooks=serialized_webhooks,
        )

        return ActorRun.model_validate(api_result)

    @_ensure_context
    async def abort(
        self,
        run_id: str,
        *,
        token: str | None = None,
        status_message: str | None = None,
        gracefully: bool | None = None,
    ) -> ActorRun:
        """Abort given Actor run on the Apify platform using the current user account.

        The user account is determined by the `APIFY_TOKEN` environment variable.

        Args:
            run_id: The ID of the Actor run to be aborted.
            token: The Apify API token to use for this request (defaults to the `APIFY_TOKEN` environment variable).
            status_message: Status message of the Actor to be set on the platform.
            gracefully: If True, the Actor run will abort gracefully. It will send `aborting` and `persistState`
                events into the run and force-stop the run after 30 seconds. It is helpful in cases where you plan
                to resurrect the run later.

        Returns:
            Info about the aborted Actor run.
        """
        client = self.new_client(token=token) if token else self.apify_client

        if status_message:
            await client.run(run_id).update(status_message=status_message)

        api_result = await client.run(run_id).abort(gracefully=gracefully)

        return ActorRun.model_validate(api_result)

    @_ensure_context
    async def call(
        self,
        actor_id: str,
        run_input: Any = None,
        *,
        token: str | None = None,
        content_type: str | None = None,
        build: str | None = None,
        memory_mbytes: int | None = None,
        timeout: timedelta | None | Literal['inherit', 'RemainingTime'] = None,
        webhooks: list[Webhook] | None = None,
        wait: timedelta | None = None,
        logger: logging.Logger | None | Literal['default'] = 'default',
    ) -> ActorRun | None:
        """Start an Actor on the Apify Platform and wait for it to finish before returning.

        It waits indefinitely, unless the wait argument is provided.

        Args:
            actor_id: The ID of the Actor to be run.
            run_input: The input to pass to the Actor run.
            token: The Apify API token to use for this request (defaults to the `APIFY_TOKEN` environment variable).
            content_type: The content type of the input.
            build: Specifies the Actor build to run. It can be either a build tag or build number. By default,
                the run uses the build specified in the default run configuration for the Actor (typically latest).
            memory_mbytes: Memory limit for the run, in megabytes. By default, the run uses a memory limit specified
                in the default run configuration for the Actor.
            timeout: Optional timeout for the run, in seconds. By default, the run uses timeout specified in
                the default run configuration for the Actor. Using `inherit` or `RemainingTime` will set timeout of the
                other Actor to the time remaining from this Actor timeout.
            webhooks: Optional webhooks (https://docs.apify.com/webhooks) associated with the Actor run, which can
                be used to receive a notification, e.g. when the Actor finished or failed. If you already have
                a webhook set up for the Actor, you do not have to add it again here.
            wait: The maximum number of seconds the server waits for the run to finish. If not provided,
                waits indefinitely.
            logger: Logger used to redirect logs from the Actor run. Using "default" literal means that a predefined
                default logger will be used. Setting `None` will disable any log propagation. Passing custom logger
                will redirect logs to the provided logger.

        Returns:
            Info about the started Actor run.
        """
        client = self.new_client(token=token) if token else self.apify_client

        if webhooks:
            serialized_webhooks = [
                hook.model_dump(by_alias=True, exclude_unset=True, exclude_defaults=True) for hook in webhooks
            ]
        else:
            serialized_webhooks = None

        if timeout in {'inherit', 'RemainingTime'}:
            if timeout == 'RemainingTime':
                warnings.warn(
                    '`RemainingTime` is deprecated and will be removed in version 4.0.0. Use `inherit` instead.',
                    DeprecationWarning,
                    stacklevel=2,
                )

            actor_call_timeout = self._get_remaining_time()
        elif timeout is None:
            actor_call_timeout = None
        elif isinstance(timeout, timedelta):
            actor_call_timeout = timeout
        else:
            raise ValueError(
                f'Invalid timeout {timeout!r}: expected `None`, `"inherit"`, `"RemainingTime"`, or a `timedelta`.'
            )

        api_result = await client.actor(actor_id).call(
            run_input=run_input,
            content_type=content_type,
            build=build,
            memory_mbytes=memory_mbytes,
            timeout_secs=int(actor_call_timeout.total_seconds()) if actor_call_timeout is not None else None,
            webhooks=serialized_webhooks,
            wait_secs=int(wait.total_seconds()) if wait is not None else None,
            logger=logger,
        )

        return ActorRun.model_validate(api_result)

    @_ensure_context
    async def call_task(
        self,
        task_id: str,
        task_input: dict | None = None,
        *,
        build: str | None = None,
        memory_mbytes: int | None = None,
        timeout: timedelta | None | Literal['inherit'] = None,
        webhooks: list[Webhook] | None = None,
        wait: timedelta | None = None,
        token: str | None = None,
    ) -> ActorRun | None:
        """Start an Actor task on the Apify Platform and wait for it to finish before returning.

        It waits indefinitely, unless the wait argument is provided.

        Note that an Actor task is a saved input configuration and options for an Actor. If you want to run an Actor
        directly rather than an Actor task, please use the `Actor.call`

        Args:
            task_id: The ID of the Actor to be run.
            task_input: Overrides the input to pass to the Actor run.
            token: The Apify API token to use for this request (defaults to the `APIFY_TOKEN` environment variable).
            content_type: The content type of the input.
            build: Specifies the Actor build to run. It can be either a build tag or build number. By default,
                the run uses the build specified in the default run configuration for the Actor (typically latest).
            memory_mbytes: Memory limit for the run, in megabytes. By default, the run uses a memory limit specified
                in the default run configuration for the Actor.
            timeout: Optional timeout for the run, in seconds. By default, the run uses timeout specified in
                the default run configuration for the Actor. Using `inherit` will set timeout of the other Actor to the
                time remaining from this Actor timeout.
            webhooks: Optional webhooks (https://docs.apify.com/webhooks) associated with the Actor run, which can
                be used to receive a notification, e.g. when the Actor finished or failed. If you already have
                a webhook set up for the Actor, you do not have to add it again here.
            wait: The maximum number of seconds the server waits for the run to finish. If not provided, waits
                indefinitely.

        Returns:
            Info about the started Actor run.
        """
        client = self.new_client(token=token) if token else self.apify_client

        if webhooks:
            serialized_webhooks = [
                hook.model_dump(by_alias=True, exclude_unset=True, exclude_defaults=True) for hook in webhooks
            ]
        else:
            serialized_webhooks = None

        if timeout == 'inherit':
            task_call_timeout = self._get_remaining_time()
        elif timeout is None:
            task_call_timeout = None
        elif isinstance(timeout, timedelta):
            task_call_timeout = timeout
        else:
            raise ValueError(f'Invalid timeout {timeout!r}: expected `None`, `"inherit"`, or a `timedelta`.')

        api_result = await client.task(task_id).call(
            task_input=task_input,
            build=build,
            memory_mbytes=memory_mbytes,
            timeout_secs=int(task_call_timeout.total_seconds()) if task_call_timeout is not None else None,
            webhooks=serialized_webhooks,
            wait_secs=int(wait.total_seconds()) if wait is not None else None,
        )

        return ActorRun.model_validate(api_result)

    @_ensure_context
    async def metamorph(
        self,
        target_actor_id: str,
        run_input: Any = None,
        *,
        target_actor_build: str | None = None,
        content_type: str | None = None,
        custom_after_sleep: timedelta | None = None,
    ) -> None:
        """Transform this Actor run to an Actor run of a different Actor.

        The platform stops the current Actor container and starts a new container with the new Actor instead. All
        the default storages are preserved, and the new input is stored under the `INPUT-METAMORPH-1` key in the same
        default key-value store.

        Args:
            target_actor_id: ID of the target Actor that the run should be transformed into
            run_input: The input to pass to the new run.
            target_actor_build: The build of the target Actor. It can be either a build tag or build number.
                By default, the run uses the build specified in the default run configuration for the target Actor
                (typically the latest build).
            content_type: The content type of the input.
            custom_after_sleep: How long to sleep for after the metamorph, to wait for the container to be stopped.
        """
        if not self.is_at_home():
            self.log.error('Actor.metamorph() is only supported when running on the Apify platform.')
            return

        if not custom_after_sleep:
            custom_after_sleep = self.configuration.metamorph_after_sleep

        # If is_at_home() is True, configuration.actor_run_id is always set
        if not self.configuration.actor_run_id:
            raise RuntimeError('actor_run_id cannot be None when running on the Apify platform.')

        await self.apify_client.run(self.configuration.actor_run_id).metamorph(
            target_actor_id=target_actor_id,
            run_input=run_input,
            target_actor_build=target_actor_build,
            content_type=content_type,
        )

        if custom_after_sleep:
            await asyncio.sleep(custom_after_sleep.total_seconds())

    @_ensure_context
    async def reboot(
        self,
        *,
        event_listeners_timeout: timedelta | None = EVENT_LISTENERS_TIMEOUT,  # noqa: ARG002
        custom_after_sleep: timedelta | None = None,
    ) -> None:
        """Internally reboot this Actor.

        The system stops the current container and starts a new one, with the same run ID and default storages.

        Args:
            event_listeners_timeout: How long should the Actor wait for Actor event listeners to finish before exiting.
            custom_after_sleep: How long to sleep for after the reboot, to wait for the container to be stopped.
        """
        if not self.is_at_home():
            self.log.error('Actor.reboot() is only supported when running on the Apify platform.')
            return

        if self._is_rebooting:
            self.log.debug('Actor is already rebooting, skipping the additional reboot call.')
            return

        self._is_rebooting = True

        if not custom_after_sleep:
            custom_after_sleep = self.configuration.metamorph_after_sleep

        # Call all the listeners for the PERSIST_STATE and MIGRATING events, and wait for them to finish.
        # PERSIST_STATE listeners are called to allow the Actor to persist its state before the reboot.
        # MIGRATING listeners are called to allow the Actor to gracefully stop in-progress tasks before the reboot.
        # Typically, crawlers are listening for the MIIGRATING event to stop processing new requests.
        # We can't just emit the events and wait for all listeners to finish,
        # because this method might be called from an event listener itself, and we would deadlock.
        persist_state_listeners = flatten(
            (self.event_manager._listeners_to_wrappers[Event.PERSIST_STATE] or {}).values()  # noqa: SLF001
        )
        migrating_listeners = flatten(
            (self.event_manager._listeners_to_wrappers[Event.MIGRATING] or {}).values()  # noqa: SLF001
        )

        await asyncio.gather(
            *[listener(EventPersistStateData(is_migrating=True)) for listener in persist_state_listeners],
            *[listener(EventMigratingData()) for listener in migrating_listeners],
        )

        if not self.configuration.actor_run_id:
            raise RuntimeError('actor_run_id cannot be None when running on the Apify platform.')

        await self.apify_client.run(self.configuration.actor_run_id).reboot()

        if custom_after_sleep:
            await asyncio.sleep(custom_after_sleep.total_seconds())

    @_ensure_context
    async def add_webhook(
        self,
        webhook: Webhook,
        *,
        ignore_ssl_errors: bool | None = None,
        do_not_retry: bool | None = None,
        idempotency_key: str | None = None,
    ) -> None:
        """Create an ad-hoc webhook for the current Actor run.

        This webhook lets you receive a notification when the Actor run finished or failed.

        Note that webhooks are only supported for Actors running on the Apify platform. When running the Actor locally,
        the function will print a warning and have no effect.

        For more information about Apify Actor webhooks, please see the [documentation](https://docs.apify.com/webhooks).

        Args:
            webhook: The webhook to be added
            ignore_ssl_errors: Whether the webhook should ignore SSL errors returned by request_url
            do_not_retry: Whether the webhook should retry sending the payload to request_url upon failure.
            idempotency_key: A unique identifier of a webhook. You can use it to ensure that you won't create
                the same webhook multiple times.

        Returns:
            The created webhook.
        """
        if not self.is_at_home():
            self.log.error('Actor.add_webhook() is only supported when running on the Apify platform.')
            return

        # If is_at_home() is True, config.actor_run_id is always set
        if not self.configuration.actor_run_id:
            raise RuntimeError('actor_run_id cannot be None when running on the Apify platform.')

        await self.apify_client.webhooks().create(
            actor_run_id=self.configuration.actor_run_id,
            event_types=webhook.event_types,
            request_url=webhook.request_url,
            payload_template=webhook.payload_template,
            ignore_ssl_errors=ignore_ssl_errors,
            do_not_retry=do_not_retry,
            idempotency_key=idempotency_key,
        )

    @_ensure_context
    async def set_status_message(
        self,
        status_message: str,
        *,
        is_terminal: bool | None = None,
    ) -> ActorRun | None:
        """Set the status message for the current Actor run.

        Args:
            status_message: The status message to set to the run.
            is_terminal: Set this flag to True if this is the final status message of the Actor run.

        Returns:
            The updated Actor run object.
        """
        if not self.is_at_home():
            title = 'Terminal status message' if is_terminal else 'Status message'
            self.log.info(f'[{title}]: {status_message}')
            return None

        # If is_at_home() is True, config.actor_run_id is always set
        if not self.configuration.actor_run_id:
            raise RuntimeError('actor_run_id cannot be None when running on the Apify platform.')

        api_result = await self.apify_client.run(self.configuration.actor_run_id).update(
            status_message=status_message, is_status_message_terminal=is_terminal
        )

        return ActorRun.model_validate(api_result)

    @_ensure_context
    async def create_proxy_configuration(
        self,
        *,
        actor_proxy_input: dict
        | None = None,  # this is the raw proxy input from the Actor run input, it is not spread or snake_cased in here
        password: str | None = None,
        groups: list[str] | None = None,
        country_code: str | None = None,
        proxy_urls: list[str | None] | None = None,
        new_url_function: _NewUrlFunction | None = None,
    ) -> ProxyConfiguration | None:
        """Create a ProxyConfiguration object with the passed proxy configuration.

        Configures connection to a proxy server with the provided options. Proxy servers are used to prevent target
        websites from blocking your crawlers based on IP address rate limits or blacklists.

        For more details and code examples, see the `ProxyConfiguration` class.

        Args:
            actor_proxy_input: Proxy configuration field from the Actor input, if input has such input field. If you
                pass this argument, all the other arguments will be inferred from it.
            password: Password for the Apify Proxy. If not provided, will use os.environ['APIFY_PROXY_PASSWORD'],
                if available.
            groups: Proxy groups which the Apify Proxy should use, if provided.
            country_code: Country which the Apify Proxy should use, if provided.
            proxy_urls: Custom proxy server URLs which should be rotated through.
            new_url_function: Function which returns a custom proxy URL to be used.

        Returns:
            ProxyConfiguration object with the passed configuration, or None, if no proxy should be used based
            on the configuration.
        """
        if actor_proxy_input is not None:
            if actor_proxy_input.get('useApifyProxy', False):
                country_code = country_code or actor_proxy_input.get('apifyProxyCountry')
                groups = groups or actor_proxy_input.get('apifyProxyGroups')
            else:
                proxy_urls = actor_proxy_input.get('proxyUrls', [])
                if not proxy_urls:
                    return None

        proxy_configuration = ProxyConfiguration(
            password=password,
            groups=groups,
            country_code=country_code,
            proxy_urls=proxy_urls,
            new_url_function=new_url_function,
            _actor_config=self.configuration,
            _apify_client=self._apify_client,
        )

        await proxy_configuration.initialize()

        return proxy_configuration

    @_ensure_context
    async def use_state(
        self,
        default_value: dict[str, JsonSerializable] | None = None,
        key: str | None = None,
        kvs_name: str | None = None,
    ) -> dict[str, JsonSerializable]:
        """Easily create and manage state values. All state values are automatically persisted.

        Values can be modified by simply using the assignment operator.

        Args:
            default_value: The default value to initialize the state if it is not already set.
            key: The key in the key-value store where the state is stored. If not provided, a default key is used.
            kvs_name: The name of the key-value store where the state is stored. If not provided, the default
                key-value store associated with the Actor run is used.

        Returns:
            The state dictionary with automatic persistence.
        """
        self._use_state_stores.add(kvs_name)
        kvs = await self.open_key_value_store(name=kvs_name)
        return await kvs.get_auto_saved_value(key or self._ACTOR_STATE_KEY, default_value)

    async def _save_actor_state(self) -> None:
        for kvs_name in self._use_state_stores:
            store = await self.open_key_value_store(name=kvs_name)
            await store.persist_autosaved_values()

    def _get_default_exit_process(self) -> bool:
        """Return False for IPython and Scrapy environments, True otherwise."""
        if is_running_in_ipython():
            self.log.debug('Running in IPython, setting default `exit_process` to False.')
            return False

        # Check if running in Scrapy by attempting to import it.
        with suppress(ImportError):
            import scrapy  # noqa: F401 PLC0415

            self.log.debug('Running in Scrapy, setting default `exit_process` to False.')
            return False

        return True

    def _get_remaining_time(self) -> timedelta | None:
        """Get time remaining from the Actor timeout. Returns `None` if not on an Apify platform."""
        if self.is_at_home() and self.configuration.timeout_at:
            return max(self.configuration.timeout_at - datetime.now(tz=timezone.utc), timedelta(0))

        self.log.warning(
            'Using `inherit` or `RemainingTime` argument is only possible when the Actor'
            ' is running on the Apify platform and when the timeout for the Actor run is set. '
            f'{self.is_at_home()=}, {self.configuration.timeout_at=}'
        )
        return None


Actor = cast('_ActorType', Proxy(_ActorType))
"""The entry point of the SDK, through which all the Actor operations should be done."""



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, get_args
from unittest.mock import Mock

import pytest
from yarl import URL

from crawlee._request import CrawleeRequestData, UserData
from crawlee._types import HttpMethod

from apify.request_loaders import ApifyRequestList
from apify.request_loaders._apify_request_list import URL_NO_COMMAS_REGEX

if TYPE_CHECKING:
    from pytest_httpserver import HTTPServer
    from werkzeug import Request, Response


@pytest.mark.parametrize(
    argnames='request_method',
    argvalues=[
        pytest.param(
            method,
            id=str(method),
        )
        for method in get_args(HttpMethod)
    ],
)
@pytest.mark.parametrize(
    argnames='optional_input',
    argvalues=[
        pytest.param({}, id='minimal'),
        pytest.param(
            {
                'payload': 'some payload',
                'userData': {'some key': 'some value'},
                'headers': {'h1': 'v1', 'h2': 'v2'},
            },
            id='all_options',
        ),
    ],
)
async def test_request_list_open_request_types(
    request_method: HttpMethod,
    optional_input: dict[str, Any],
) -> None:
    """Test proper request list generation from both minimal and full inputs for all method types for simple input."""
    minimal_request_dict_input = {
        'url': 'https://www.abc.com',
        'method': request_method,
    }
    request_dict_input = {**minimal_request_dict_input, **optional_input}

    request_list = await ApifyRequestList.open(request_list_sources_input=[request_dict_input])
    assert not await request_list.is_empty()

    request = await request_list.fetch_next_request()
    assert request is not None
    assert await request_list.is_empty(), 'Request list should be empty after fetching all requests'

    assert request.method == request_dict_input['method']
    assert request.url == request_dict_input['url']
    assert request.payload == request_dict_input.get('payload', '').encode('utf-8')
    expected_user_data = UserData()
    # `crawlee_data` must be present in user_data
    expected_user_data.crawlee_data = CrawleeRequestData()
    if 'userData' in optional_input:
        for key, value in optional_input['userData'].items():
            expected_user_data[key] = value
    assert request.user_data == expected_user_data
    assert request.headers.root == optional_input.get('headers', {})


async def test_request_list_open_from_url_correctly_send_requests(httpserver: HTTPServer) -> None:
    """Test that requests are sent to expected urls."""
    request_list_sources_input: list[dict[str, Any]] = [
        {
            'requestsFromUrl': httpserver.url_for('/file.txt'),
            'method': 'GET',
        },
        {
            'requestsFromUrl': httpserver.url_for('/file2'),
            'method': 'PUT',
        },
        {
            'requestsFromUrl': httpserver.url_for('/something'),
            'method': 'POST',
            'headers': {'key': 'value'},
            'payload': 'some_payload',
            'userData': {'another_key': 'another_value'},
        },
    ]

    routes: dict[str, Mock] = {}

    def request_handler(request: Request, response: Response) -> Response:
        routes[request.url]()
        return response

    for entry in request_list_sources_input:
        path = str(URL(entry['requestsFromUrl']).path)
        httpserver.expect_oneshot_request(path).with_post_hook(request_handler).respond_with_data(status=200)
        routes[entry['requestsFromUrl']] = Mock()

    await ApifyRequestList.open(request_list_sources_input=request_list_sources_input)

    assert len(routes) == len(request_list_sources_input)

    for entity in request_list_sources_input:
        entity_url = entity['requestsFromUrl']
        assert entity_url in routes
        assert routes[entity_url].called


async def test_request_list_open_from_url(httpserver: HTTPServer) -> None:
    """Test that create_request_list is correctly reading urls from remote url sources and also from simple input."""
    expected_simple_url = 'https://www.someurl.com'
    expected_remote_urls_1 = {'http://www.something.com', 'https://www.somethingelse.com', 'http://www.bla.net'}
    expected_remote_urls_2 = {'http://www.ok.com', 'https://www.true-positive.com'}
    expected_urls = expected_remote_urls_1 | expected_remote_urls_2 | {expected_simple_url}

    @dataclass
    class MockedUrlInfo:
        url: str
        response_text: str

    mocked_urls = (
        MockedUrlInfo(
            httpserver.url_for('/file.txt'),
            'blablabla{} more blablabla{} , even more blablabla. {} '.format(*expected_remote_urls_1),
        ),
        MockedUrlInfo(
            httpserver.url_for('/file2'),
            'some stuff{} more stuff{} www.false_positive.com'.format(*expected_remote_urls_2),
        ),
    )

    request_list_sources_input = [
        {
            'requestsFromUrl': mocked_urls[0].url,
            'method': 'GET',
        },
        {'url': expected_simple_url, 'method': 'GET'},
        {
            'requestsFromUrl': mocked_urls[1].url,
            'method': 'GET',
        },
    ]
    for mocked_url in mocked_urls:
        path = str(URL(mocked_url.url).path)
        httpserver.expect_oneshot_request(path).respond_with_data(status=200, response_data=mocked_url.response_text)

    request_list = await ApifyRequestList.open(request_list_sources_input=request_list_sources_input)
    generated_requests = []
    while request := await request_list.fetch_next_request():
        generated_requests.append(request)

    # Check correctly created requests' urls in request list
    assert {generated_request.url for generated_request in generated_requests} == expected_urls


async def test_request_list_open_from_url_additional_inputs(httpserver: HTTPServer) -> None:
    """Test that all generated request properties are correctly populated from input values."""
    expected_url = 'https://www.someurl.com'
    example_start_url_input: dict[str, Any] = {
        'requestsFromUrl': httpserver.url_for('/file.txt'),
        'method': 'POST',
        'headers': {'key': 'value'},
        'payload': 'some_payload',
        'userData': {'another_key': 'another_value'},
    }
    httpserver.expect_oneshot_request('/file.txt').respond_with_data(status=200, response_data=expected_url)

    request_list = await ApifyRequestList.open(request_list_sources_input=[example_start_url_input])
    request = await request_list.fetch_next_request()
    # Check all properties correctly created for request
    assert request
    assert request.url == expected_url
    assert request.method == example_start_url_input['method']
    assert request.headers.root == example_start_url_input['headers']
    assert request.payload == str(example_start_url_input['payload']).encode('utf-8')
    expected_user_data = UserData()
    # `crawlee_data` must be present in user_data
    expected_user_data.crawlee_data = CrawleeRequestData()
    for key, value in example_start_url_input['userData'].items():
        expected_user_data[key] = value
    assert request.user_data == expected_user_data


async def test_request_list_open_name() -> None:
    name = 'some_name'
    request_list = await ApifyRequestList.open(name=name)
    assert request_list.name == name


@pytest.mark.parametrize(
    argnames='true_positive',
    argvalues=[
        pytest.param('http://www.something.com', id='standard_http_with_www'),
        pytest.param('https://www.something.net', id='standard_https_with_www'),
        pytest.param('http://nowww.cz', id='http_no_www'),
        pytest.param('https://with-hyphen.com', id='https_with_hyphen'),
        pytest.param('http://number1.com', id='http_with_number_in_domain'),
        pytest.param('http://www.number.123.abc', id='http_with_subdomains_and_numbers'),
        pytest.param('http://many.dots.com', id='http_with_multiple_subdomains'),
        pytest.param('http://a.com', id='http_short_domain'),
        pytest.param('http://www.something.com/somethignelse', id='http_with_path_no_extension'),
        pytest.param('http://www.something.com/somethignelse.txt', id='http_with_path_and_extension'),
        pytest.param('http://non-english-chars-áíéåü.com', id='http_with_non_english_chars'),
        pytest.param('http://www.port.com:1234', id='http_with_port'),
        pytest.param('http://username:password@something.else.com', id='http_with_authentication'),
    ],
)
def test_url_no_commas_regex_true_positives(true_positive: str) -> None:
    example_string = f'Some text {true_positive} some more text'
    matches = list(re.finditer(URL_NO_COMMAS_REGEX, example_string))
    assert len(matches) == 1
    assert matches[0].group(0) == true_positive


@pytest.mark.parametrize(
    argnames='false_positive',
    argvalues=[
        pytest.param('http://www.a', id='invalid_domain_single_level'),
        pytest.param('http://a', id='invalid_domain_no_tld'),
        pytest.param('http://a.a', id='invalid_domain_short_tld'),
        pytest.param('http://123.456', id='invalid_numeric_domain'),
        pytest.param('www.something.com', id='missing_protocol'),
        pytest.param('http:www.something.com', id='missing_slashes'),
    ],
)
def test_url_no_commas_regex_false_positives(false_positive: str) -> None:
    example_string = f'Some text {false_positive} some more text'
    matches = list(re.findall(URL_NO_COMMAS_REGEX, example_string))
    assert len(matches) == 0


def test_url_no_commas_regex_multi_line() -> None:
    true_positives = ('http://www.something.com', 'http://www.else.com')
    example_string = 'Some text {} some more text \n Some new line text {} ...'.format(*true_positives)
    matches = list(re.finditer(URL_NO_COMMAS_REGEX, example_string))
    assert len(matches) == 2
    assert {match.group(0) for match in matches} == set(true_positives)



from __future__ import annotations

from datetime import datetime
from typing import Annotated, Any, Literal

from pydantic import BaseModel, Field

from crawlee.events._types import (
    Event,
    EventAbortingData,
    EventExitData,
    EventMigratingData,
    EventPersistStateData,
    EventSystemInfoData,
)

from apify._utils import docs_group


@docs_group('Event data')
class SystemInfoEventData(BaseModel):
    mem_avg_bytes: Annotated[float, Field(alias='memAvgBytes')]
    mem_current_bytes: Annotated[float, Field(alias='memCurrentBytes')]
    mem_max_bytes: Annotated[float, Field(alias='memMaxBytes')]
    cpu_avg_usage: Annotated[float, Field(alias='cpuAvgUsage')]
    cpu_max_usage: Annotated[float, Field(alias='cpuMaxUsage')]
    cpu_current_usage: Annotated[float, Field(alias='cpuCurrentUsage')]
    is_cpu_overloaded: Annotated[bool, Field(alias='isCpuOverloaded')]
    created_at: Annotated[datetime, Field(alias='createdAt')]

    def to_crawlee_format(self, dedicated_cpus: float) -> EventSystemInfoData:
        return EventSystemInfoData.model_validate(
            {
                'cpu_info': {
                    'used_ratio': (self.cpu_current_usage / 100) / dedicated_cpus,
                    'created_at': self.created_at,
                },
                'memory_info': {
                    'total_size': self.mem_max_bytes,
                    'current_size': self.mem_current_bytes,
                    'created_at': self.created_at,
                },
            }
        )


@docs_group('Events')
class PersistStateEvent(BaseModel):
    name: Literal[Event.PERSIST_STATE]
    data: Annotated[EventPersistStateData, Field(default_factory=lambda: EventPersistStateData(is_migrating=False))]


@docs_group('Events')
class SystemInfoEvent(BaseModel):
    name: Literal[Event.SYSTEM_INFO]
    data: SystemInfoEventData


@docs_group('Events')
class MigratingEvent(BaseModel):
    name: Literal[Event.MIGRATING]
    data: Annotated[EventMigratingData, Field(default_factory=EventMigratingData)]


@docs_group('Events')
class AbortingEvent(BaseModel):
    name: Literal[Event.ABORTING]
    data: Annotated[EventAbortingData, Field(default_factory=EventAbortingData)]


@docs_group('Events')
class ExitEvent(BaseModel):
    name: Literal[Event.EXIT]
    data: Annotated[EventExitData, Field(default_factory=EventExitData)]


@docs_group('Events')
class EventWithoutData(BaseModel):
    name: Literal[
        Event.SESSION_RETIRED,
        Event.BROWSER_LAUNCHED,
        Event.BROWSER_RETIRED,
        Event.BROWSER_CLOSED,
        Event.PAGE_CREATED,
        Event.PAGE_CLOSED,
    ]
    data: Any = None


@docs_group('Events')
class DeprecatedEvent(BaseModel):
    name: Literal['cpuInfo']
    data: Annotated[dict[str, Any], Field(default_factory=dict)]


@docs_group('Events')
class UnknownEvent(BaseModel):
    name: str
    data: Annotated[dict[str, Any], Field(default_factory=dict)]


EventMessage = PersistStateEvent | SystemInfoEvent | MigratingEvent | AbortingEvent | ExitEvent | EventWithoutData



from __future__ import annotations

from crawlee.crawlers import BeautifulSoupCrawler, BeautifulSoupCrawlingContext

from apify import Actor


async def main() -> None:
    async with Actor:
        pages_visited: list[str] = []
        crawler = BeautifulSoupCrawler(max_crawl_depth=2)

        @crawler.router.default_handler
        async def handler(context: BeautifulSoupCrawlingContext) -> None:
            pages_visited.append(context.request.url)
            await context.enqueue_links()

            if '/products/' in context.request.url:
                name_tag = context.soup.find('h1')
                price_tag = context.soup.find('span', class_='price')
                desc_tag = context.soup.find('p', class_='description')
                if name_tag:
                    await context.push_data(
                        {
                            'url': context.request.url,
                            'name': name_tag.get_text(strip=True),
                            'price': price_tag.get_text(strip=True) if price_tag else '',
                            'description': desc_tag.get_text(strip=True) if desc_tag else '',
                        }
                    )

        await crawler.run(['http://localhost:8080/'])

        await Actor.set_value(
            'CRAWLER_RESULT',
            {
                'pages_visited_count': len(pages_visited),
                'crawler_type': 'BeautifulSoupCrawler',
            },
        )



from __future__ import annotations

import re
from base64 import b64encode
from hashlib import sha256
from typing import TYPE_CHECKING

from crawlee._utils.crypto import compute_short_hash

if TYPE_CHECKING:
    from apify import Configuration


def unique_key_to_request_id(unique_key: str, *, request_id_length: int = 15) -> str:
    """Generate a deterministic request ID based on a unique key.

    Args:
        unique_key: The unique key to convert into a request ID.
        request_id_length: The length of the request ID.

    Returns:
        A URL-safe, truncated request ID based on the unique key.
    """
    # Encode the unique key and compute its SHA-256 hash
    hashed_key = sha256(unique_key.encode('utf-8')).digest()

    # Encode the hash in base64 and decode it to get a string
    base64_encoded = b64encode(hashed_key).decode('utf-8')

    # Remove characters that are not URL-safe ('+', '/', or '=')
    url_safe_key = re.sub(r'(\+|\/|=)', '', base64_encoded)

    # Truncate the key to the desired length
    return url_safe_key[:request_id_length]


def hash_api_base_url_and_token(configuration: Configuration) -> str:
    """Hash configuration.api_public_base_url and configuration.token in deterministic way."""
    if configuration.api_public_base_url is None or configuration.token is None:
        raise ValueError("'Configuration.api_public_base_url' and 'Configuration.token' must be set.")
    return compute_short_hash(f'{configuration.api_public_base_url}{configuration.token}'.encode())



from apify import Actor


async def main() -> None:
    async with Actor:
        await Actor.set_status_message('Here we go!')
        # Do some work...
        await Actor.set_status_message('So far so good...')
        # Do some more work...
        await Actor.set_status_message('Steady as she goes...')
        # Do even more work...
        await Actor.set_status_message('Almost there...')
        # Finish the job
        await Actor.set_status_message('Phew! That was not that hard!')



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Coroutine


def run_scrapy_actor(coro: Coroutine) -> None:
    """Start Twisted's reactor and execute the provided Actor coroutine.

    This function installs Twisted's asyncio-compatible reactor, then initiates it and runs the given asyncio
    coroutine (typically the Actor's main) by converting it to a Deferred. This bridges the asyncio and Twisted
    event loops, enabling the Apify and Scrapy integration to work together.
    """
    from scrapy.utils.reactor import install_reactor  # noqa: PLC0415
    from twisted.internet.error import ReactorAlreadyInstalledError  # noqa: PLC0415

    try:
        install_reactor('twisted.internet.asyncioreactor.AsyncioSelectorReactor')
    except ReactorAlreadyInstalledError:
        from twisted.internet import reactor as installed_reactor  # noqa: PLC0415

        reactor_cls = type(installed_reactor)
        if (
            reactor_cls.__module__ != 'twisted.internet.asyncioreactor'
            or reactor_cls.__name__ != 'AsyncioSelectorReactor'
        ):
            raise RuntimeError(
                'A Twisted reactor is already installed and it is not AsyncioSelectorReactor. '
                'Make sure that run_scrapy_actor() is called before importing any Scrapy or Twisted '
                'modules that install a reactor (e.g. scrapy.crawler).'
            ) from None

    from twisted.internet.defer import Deferred  # noqa: PLC0415
    from twisted.internet.task import react  # noqa: PLC0415

    react(lambda _reactor: Deferred.fromFuture(asyncio.ensure_future(coro)))



import json
from decimal import Decimal
from pathlib import Path

import pytest

from crawlee import Request, service_locator
from crawlee._types import BasicCrawlingContext
from crawlee.configuration import Configuration as CrawleeConfiguration
from crawlee.crawlers import BasicCrawler
from crawlee.errors import ServiceConflictError

from apify import Actor
from apify import Configuration as ApifyConfiguration
from apify.storage_clients._smart_apify._storage_client import SmartApifyStorageClient


@pytest.mark.parametrize(
    ('is_at_home', 'disable_browser_sandbox_in', 'disable_browser_sandbox_out'),
    [
        (False, False, False),
        (False, True, True),
        (True, False, True),
        (True, True, True),
    ],
)
def test_disable_browser_sandbox(
    *, is_at_home: bool, disable_browser_sandbox_in: bool, disable_browser_sandbox_out: bool
) -> None:
    assert (
        ApifyConfiguration(
            is_at_home=is_at_home, disable_browser_sandbox=disable_browser_sandbox_in
        ).disable_browser_sandbox
        == disable_browser_sandbox_out
    )


async def test_existing_apify_config_respected_by_actor() -> None:
    """Set Apify Configuration in service_locator and verify that Actor respects it."""
    max_used_cpu_ratio = 0.123456  # Some unique value to verify configuration
    apify_config = ApifyConfiguration(max_used_cpu_ratio=max_used_cpu_ratio)
    service_locator.set_configuration(apify_config)
    async with Actor:
        pass

    returned_config = service_locator.get_configuration()
    assert returned_config is apify_config


async def test_existing_crawlee_config_respected_by_actor() -> None:
    """Set Crawlee Configuration in service_locator and verify that Actor respects it."""
    max_used_cpu_ratio = 0.123456  # Some unique value to verify configuration
    crawlee_config = CrawleeConfiguration(max_used_cpu_ratio=max_used_cpu_ratio)
    service_locator.set_configuration(crawlee_config)
    async with Actor:
        pass

    assert Actor.configuration is not crawlee_config
    assert isinstance(Actor.configuration, ApifyConfiguration)
    # Make sure the Crawlee Configuration was used to create ApifyConfiguration in Actor
    assert Actor.configuration.max_used_cpu_ratio == max_used_cpu_ratio


async def test_existing_apify_config_throws_error_when_set_in_actor() -> None:
    """Test that passing explicit configuration to actor after service locator configuration was already set,
    raises exception."""
    service_locator.set_configuration(ApifyConfiguration())
    with pytest.raises(ServiceConflictError):
        async with Actor(configuration=ApifyConfiguration()):
            pass


async def test_setting_config_after_actor_raises_exception() -> None:
    """Test that setting configuration in service locator after actor was created raises an exception."""
    async with Actor():
        with pytest.raises(ServiceConflictError):
            service_locator.set_configuration(ApifyConfiguration())


async def test_actor_using_input_configuration() -> None:
    """Test that configuration passed to Actor is stored in the service locator."""
    apify_config = ApifyConfiguration()
    async with Actor(configuration=apify_config):
        pass

    assert service_locator.get_configuration() is apify_config


async def test_crawler_implicit_configuration_through_actor() -> None:
    """Test that crawler uses Actor configuration unless explicit configuration was passed to it."""
    apify_config = ApifyConfiguration()
    async with Actor(configuration=apify_config):
        crawler = BasicCrawler()

    assert crawler._service_locator.get_configuration() is apify_config
    assert service_locator.get_configuration() is apify_config


async def test_crawler_implicit_configuration() -> None:
    """Test that crawler and Actor use implicit service_locator based configuration unless explicit configuration
    was passed to them."""
    async with Actor():
        assert Actor.configuration is service_locator.get_configuration()
        crawler = BasicCrawler()

    assert Actor.configuration is service_locator.get_configuration()
    assert Actor.configuration is crawler._service_locator.get_configuration()


async def test_crawler_implicit_local_storage() -> None:
    """Test that crawler and Actor use implicit ApifyFileSystemStorageClient."""
    async with Actor():
        crawler = BasicCrawler()

    assert isinstance(service_locator.get_storage_client(), SmartApifyStorageClient)
    assert isinstance(crawler._service_locator.get_storage_client(), SmartApifyStorageClient)


async def test_crawlers_own_configuration(tmp_path: Path) -> None:
    """Test that crawlers can use own configurations without crashing."""
    config_actor = ApifyConfiguration()
    dir_1 = tmp_path / 'dir_1'
    dir_2 = tmp_path / 'dir_2'
    config_crawler_1 = ApifyConfiguration()
    config_actor.storage_dir = str(dir_1)
    config_crawler_2 = ApifyConfiguration()
    config_crawler_2.storage_dir = str(dir_2)

    async with Actor(configuration=config_actor):

        async def request_handler(context: BasicCrawlingContext) -> None:
            Actor.log.info(f'Processing: {context.request.url}')

        crawler_1 = BasicCrawler(configuration=config_crawler_1, request_handler=request_handler)
        crawler_2 = BasicCrawler(configuration=config_crawler_2, request_handler=request_handler)
        await crawler_1.add_requests([Request.from_url(url='http://example.com/1')])
        await crawler_2.add_requests(
            [Request.from_url(url='http://example.com/2'), Request.from_url(url='http://example.com/3')]
        )

        await crawler_1.run()
        await crawler_2.run()

    assert service_locator.get_configuration() is config_actor
    assert crawler_1._service_locator.get_configuration() is config_crawler_1
    assert crawler_2._service_locator.get_configuration() is config_crawler_2

    assert crawler_1.statistics.state.requests_total == 1
    assert crawler_2.statistics.state.requests_total == 2


async def test_crawler_global_configuration() -> None:
    """Test that crawler and Actor use service_locator based configuration unless explicit configuration
    was passed to them."""
    config_global = ApifyConfiguration()
    service_locator.set_configuration(config_global)

    async with Actor():
        crawler = BasicCrawler()

    assert service_locator.get_configuration() is config_global
    assert crawler._service_locator.get_configuration() is config_global


async def test_crawler_uses_implicit_apify_config() -> None:
    """Test that Actor is using implicit ApifyConfiguration in Actor context."""
    async with Actor:
        assert isinstance(Actor.configuration, ApifyConfiguration)


async def test_storages_retrieved_is_different_with_different_config(tmp_path: Path) -> None:
    """Test that retrieving storage depends on used configuration."""
    dir_1 = tmp_path / 'dir_1'
    dir_2 = tmp_path / 'dir_2'
    config_actor = ApifyConfiguration()
    config_actor.storage_dir = str(dir_1)
    config_crawler = ApifyConfiguration()
    config_crawler.storage_dir = str(dir_2)

    async with Actor(configuration=config_actor):
        actor_kvs = await Actor.open_key_value_store()
        actor_dataset = await Actor.open_dataset()
        actor_rq = await Actor.open_request_queue()

        crawler = BasicCrawler(configuration=config_crawler)
        crawler_kvs = await crawler.get_key_value_store()
        crawler_dataset = await crawler.get_dataset()
        crawler_rq = await crawler.get_request_manager()

    assert actor_kvs is not crawler_kvs
    assert actor_dataset is not crawler_dataset
    assert actor_rq is not crawler_rq


async def test_storages_retrieved_is_same_with_equivalent_config() -> None:
    """Test that retrieving storage depends on used configuration. If two equivalent configuration(even if they are
    different instances) are used it returns same storage."""
    config_actor = ApifyConfiguration()
    config_crawler = ApifyConfiguration()

    async with Actor(configuration=config_actor):
        actor_kvs = await Actor.open_key_value_store()
        actor_dataset = await Actor.open_dataset()
        actor_rq = await Actor.open_request_queue()

        crawler = BasicCrawler(configuration=config_crawler)
        crawler_kvs = await crawler.get_key_value_store()
        crawler_dataset = await crawler.get_dataset()
        crawler_rq = await crawler.get_request_manager()

    assert actor_kvs is crawler_kvs
    assert actor_dataset is crawler_dataset
    assert actor_rq is crawler_rq


async def test_storages_retrieved_is_same_with_same_config() -> None:
    """Test that retrieving storage is same if same configuration is used."""
    async with Actor():
        actor_kvs = await Actor.open_key_value_store()
        actor_dataset = await Actor.open_dataset()
        actor_rq = await Actor.open_request_queue()

        crawler = BasicCrawler()
        crawler_kvs = await crawler.get_key_value_store()
        crawler_dataset = await crawler.get_dataset()
        crawler_rq = await crawler.get_request_manager()

    assert actor_kvs is crawler_kvs
    assert actor_dataset is crawler_dataset
    assert actor_rq is crawler_rq


def test_apify_configuration_is_always_used(caplog: pytest.LogCaptureFixture) -> None:
    """Set Crawlee Configuration in Actor and verify that Apify Configuration is used with warning."""
    max_used_cpu_ratio = 0.123456  # Some unique value to verify configuration

    service_locator.set_configuration(CrawleeConfiguration(max_used_cpu_ratio=max_used_cpu_ratio))

    assert Actor.configuration.max_used_cpu_ratio == max_used_cpu_ratio
    assert isinstance(Actor.configuration, ApifyConfiguration)
    assert (
        'Non Apify Configuration is set in the `service_locator` in the SDK context. '
        'It is recommended to set `apify.Configuration` explicitly as early as possible by using '
        'service_locator.set_configuration'
    ) in caplog.messages


def test_env_vars_populate_correctly(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test that configuration values are populated from environment variables."""
    monkeypatch.setenv('APIFY_TOKEN', 'my-test-token')
    monkeypatch.setenv('APIFY_ACTOR_ID', 'actor-123')
    monkeypatch.setenv('APIFY_ACT_RUN_ID', 'run-456')
    monkeypatch.setenv('APIFY_IS_AT_HOME', '1')
    monkeypatch.setenv('APIFY_API_BASE_URL', 'https://custom-api.apify.com')
    config = ApifyConfiguration()
    assert config.token == 'my-test-token'
    assert config.actor_id == 'actor-123'
    assert config.actor_run_id == 'run-456'
    assert config.is_at_home is True
    assert config.api_base_url == 'https://custom-api.apify.com'


def test_default_values() -> None:
    """Test that default values are set correctly."""
    config = ApifyConfiguration()
    assert config.is_at_home is False
    assert config.api_base_url == 'https://api.apify.com'
    assert config.api_public_base_url == 'https://api.apify.com'
    assert config.proxy_hostname == 'proxy.apify.com'
    assert config.proxy_port == 8000
    assert config.input_key == 'INPUT'
    assert config.token is None
    assert config.actor_id is None
    assert config.actor_run_id is None
    assert config.max_total_charge_usd is None
    assert config.test_pay_per_event is False


def test_max_paid_dataset_items_zero_is_preserved(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test that max_paid_dataset_items=0 is not treated as falsy and converted to None."""
    monkeypatch.setenv('ACTOR_MAX_PAID_DATASET_ITEMS', '0')
    config = ApifyConfiguration()
    assert config.max_paid_dataset_items == 0


def test_max_total_charge_usd_zero_is_preserved(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test that max_total_charge_usd=0 is not treated as falsy and converted to None."""
    monkeypatch.setenv('ACTOR_MAX_TOTAL_CHARGE_USD', '0')
    config = ApifyConfiguration()
    assert config.max_total_charge_usd == Decimal(0)


def test_max_paid_dataset_items_empty_string_becomes_none(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test that an empty env var for max_paid_dataset_items is converted to None."""
    monkeypatch.setenv('ACTOR_MAX_PAID_DATASET_ITEMS', '')
    config = ApifyConfiguration()
    assert config.max_paid_dataset_items is None


def test_max_total_charge_usd_empty_string_becomes_none(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test that an empty env var for max_total_charge_usd is converted to None."""
    monkeypatch.setenv('ACTOR_MAX_TOTAL_CHARGE_USD', '')
    config = ApifyConfiguration()
    assert config.max_total_charge_usd is None


def test_max_total_charge_usd_decimal_parsing(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test that max_total_charge_usd is parsed as Decimal from env var."""
    from decimal import Decimal

    monkeypatch.setenv('ACTOR_MAX_TOTAL_CHARGE_USD', '42.50')
    config = ApifyConfiguration()
    assert config.max_total_charge_usd == Decimal('42.50')
    assert isinstance(config.max_total_charge_usd, Decimal)


def test_actor_pricing_info_from_json_env_var(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test that actor_pricing_info is parsed from JSON env var."""

    pricing_json = json.dumps(
        {
            'pricingModel': 'PAY_PER_EVENT',
            'pricingPerEvent': {
                'actorChargeEvents': {'search': {'eventPriceUsd': '0.01', 'eventTitle': 'Search event'}}
            },
        }
    )
    monkeypatch.setenv('APIFY_ACTOR_PRICING_INFO', pricing_json)
    config = ApifyConfiguration()
    assert config.actor_pricing_info is not None
    assert config.actor_pricing_info.pricing_model == 'PAY_PER_EVENT'


def test_actor_storage_json_env_var(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test that actor_storages_json is parsed from JSON env var."""
    datasets = {'default': 'default_dataset_id', 'custom': 'custom_dataset_id'}
    request_queues = {'default': 'default_request_queue_id', 'custom': 'custom_request_queue_id'}
    key_value_stores = {'default': 'default_key_value_store_id', 'custom': 'custom_key_value_store_id'}

    actor_storages_json = json.dumps(
        {
            'datasets': datasets,
            'requestQueues': request_queues,
            'keyValueStores': key_value_stores,
        }
    )
    monkeypatch.setenv('ACTOR_STORAGES_JSON', actor_storages_json)
    config = ApifyConfiguration()
    assert config.actor_storages
    assert config.actor_storages['datasets'] == datasets
    assert config.actor_storages['request_queues'] == request_queues
    assert config.actor_storages['key_value_stores'] == key_value_stores



from __future__ import annotations

import base64

import pytest

from apify._crypto import (
    _load_public_key,
    create_hmac_signature,
    crypto_random_object_id,
    encode_base62,
    load_private_key,
    private_decrypt,
    public_encrypt,
)

# NOTE: Uses the same keys as in:
# https://github.com/apify/apify-shared-js/blob/master/test/crypto.test.ts
PRIVATE_KEY_PEM_BASE64 = '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'  # noqa: E501
PRIVATE_KEY_PASSWORD = 'pwd1234'
PUBLIC_KEY_PEM_BASE64 = 'LS0tLS1CRUdJTiBQVUJMSUMgS0VZLS0tLS0KTUlJQklqQU5CZ2txaGtpRzl3MEJBUUVGQUFPQ0FROEFNSUlCQ2dLQ0FRRUF0dis3NlNXbklhOFFKWC94RUQxRQpYdnBBQmE3ajBnQnVYenJNUU5adjhtTW1RU0t2VUF0TmpOL2xacUZpQ0haZUQxU2VDcGV1MnFHTm5XbGRxNkhUCnh5cXJpTVZEbFNKaFBNT09QSENISVNVdFI4Tk5lR1Y1MU0wYkxJcENabHcyTU9GUjdqdENWejVqZFRpZ1NvYTIKQWxrRUlRZWQ4UVlDKzk1aGJoOHk5bGcwQ0JxdEdWN1FvMFZQR2xKQ0hGaWNuaWxLVFFZay9MZzkwWVFnUElPbwozbUppeFl5bWFGNmlMZTVXNzg1M0VHWUVFVWdlWmNaZFNjaGVBMEdBMGpRSFVTdnYvMEZjay9adkZNZURJOTVsCmJVQ0JoQjFDbFg4OG4wZUhzUmdWZE5vK0NLMDI4T2IvZTZTK1JLK09VaHlFRVdPTi90alVMdGhJdTJkQWtGcmkKOFFJREFRQUIKLS0tLS1FTkQgUFVCTElDIEtFWS0tLS0tCg=='  # noqa: E501
PRIVATE_KEY = load_private_key(PRIVATE_KEY_PEM_BASE64, PRIVATE_KEY_PASSWORD)
PUBLIC_KEY = _load_public_key(PUBLIC_KEY_PEM_BASE64)


def test_encrypt_decrypt_various_strings() -> None:
    for value in [
        crypto_random_object_id(10),
        '👍',
        '!',
        '@',
        '#',
        '$',
        '%',
        '^',
        '&',
        '*',
        '(',
        ')',
        '-',
        '_',
        '=',
        '+',
        '[',
        ']',
        '{',
        '}',
        '|',
        ';',
        ':',
        '"',
        "'",
        ',',
        '.',
        '<',
        '>',
        '?',
        '/',
        '~',
    ]:
        encrypted = public_encrypt(value, public_key=PUBLIC_KEY)
        decrypted_value = private_decrypt(**encrypted, private_key=PRIVATE_KEY)
        assert decrypted_value == value


def test_decryption_fails_with_invalid_password() -> None:
    test_value = 'test'
    encrypted = public_encrypt(test_value, public_key=PUBLIC_KEY)
    encrypted['encrypted_password'] = base64.b64encode(b'invalid_password').decode('utf-8')

    with pytest.raises(ValueError, match=r'Ciphertext length must be equal to key size.'):
        private_decrypt(**encrypted, private_key=PRIVATE_KEY)


def test_decryption_fails_with_manipulated_cipher() -> None:
    test_value = 'test2'
    encrypted = public_encrypt(test_value, public_key=PUBLIC_KEY)
    encrypted['encrypted_value'] = base64.b64encode(
        b'invalid_cipher' + base64.b64decode(encrypted['encrypted_value'].encode('utf-8')),
    ).decode('utf-8')

    with pytest.raises(ValueError, match=r'Decryption failed, malformed encrypted value or password.'):
        private_decrypt(**encrypted, private_key=PRIVATE_KEY)


def test_same_value_produces_different_cipher_each_time() -> None:
    test_value = 'test3'
    encrypted1 = public_encrypt(test_value, public_key=PUBLIC_KEY)
    encrypted2 = public_encrypt(test_value, public_key=PUBLIC_KEY)
    assert encrypted1['encrypted_value'] != encrypted2['encrypted_value']


# Check if the method is compatible with js version of the same method in:
# https://github.com/apify/apify-shared-js/blob/master/packages/utilities/src/crypto.ts
def test_private_decrypt_with_node_js_encrypted_value() -> None:
    value = 'encrypted_with_node_js'
    # This was encrypted with nodejs version of the same method.
    encrypted_value_with_node_js = {
        'encrypted_password': 'lw0ez64/T1UcCQMLfhucZ6VIfMcf/TKni7PmXlL/ZRA4nmdGYz7/YQUzGWzKbLChrpqbG21DHxPIubUIQFDFE1ASkLvoSd0Ks8/wjKHMyhp+hsg5aSh9EZK6pBFpp6FeHoinV80+UURTvJuSVbWd1Orw5Frl41taP6RK3uNJlXikmgs8Xc7mShFEENgkz6y9+Pbe7jpcKkaJ2U/h7FN0eNON189kNFYVuAE1n2N6C3Q7dFnjl2e1btqErvg5Vu7ZS4BbX3wgC2qLYySGnqI3BNI5VGhAnncnQcjHb+85qG+LKoPekgY9I0s0kGMxiz/bmy1mYm9O+Lj1mbVUr7BDjQ==',  # noqa: E501
        'encrypted_value': 'k8nkZDCi0hRfBc0RRefxeSHeGV0X60N03VCrhRhENKXBjrF/tEg=',
    }
    decrypted_value = private_decrypt(
        **encrypted_value_with_node_js,
        private_key=PRIVATE_KEY,
    )

    assert decrypted_value == value


def test_crypto_random_object_id_length_and_charset() -> None:
    assert len(crypto_random_object_id()) == 17
    assert len(crypto_random_object_id(5)) == 5
    long_random_object_id = crypto_random_object_id(1000)
    for char in long_random_object_id:
        assert char in 'abcdefghijklmnopqrstuvwxyzABCEDFGHIJKLMNOPQRSTUVWXYZ0123456789'


@pytest.mark.parametrize(('test_input', 'expected'), [(0, '0'), (10, 'a'), (999999999, '15FTGf')])
def test_encode_base62(test_input: int, expected: str) -> None:
    assert encode_base62(test_input) == expected


# This test ensures compatibility with the JavaScript version of the same method.
# https://github.com/apify/apify-shared-js/blob/master/packages/utilities/src/hmac.ts
def test_create_valid_hmac_signature() -> None:
    # This test uses the same secret key and message as in JS tests.
    secret_key = 'hmac-secret-key'
    message = 'hmac-message-to-be-authenticated'
    assert create_hmac_signature(secret_key, message) == 'pcVagAsudj8dFqdlg7mG'


def test_create_same_hmac() -> None:
    # This test uses the same secret key and message as in JS tests.
    secret_key = 'hmac-same-secret-key'
    message = 'hmac-same-message-to-be-authenticated'
    assert create_hmac_signature(secret_key, message) == 'FYMcmTIm3idXqleF1Sw5'


def test_encrypt_decrypt_empty_string() -> None:
    """Test that encrypting and decrypting an empty string works correctly."""
    encrypted = public_encrypt('', public_key=PUBLIC_KEY)
    decrypted = private_decrypt(**encrypted, private_key=PRIVATE_KEY)
    assert decrypted == ''


def test_encrypt_decrypt_very_long_string() -> None:
    """Test that encrypting and decrypting a very long string works correctly."""
    long_string = 'A' * 10000
    encrypted = public_encrypt(long_string, public_key=PUBLIC_KEY)
    decrypted = private_decrypt(**encrypted, private_key=PRIVATE_KEY)
    assert decrypted == long_string


def test_hmac_with_empty_message() -> None:
    """Test HMAC signature generation with an empty message."""
    signature = create_hmac_signature('some-key', '')
    assert isinstance(signature, str)
    assert len(signature) > 0
    # Same key and empty message should always produce the same signature
    assert create_hmac_signature('some-key', '') == signature


def test_encode_base62_large_number() -> None:
    """Test encode_base62 with a large number."""
    result = encode_base62(2**64)
    assert isinstance(result, str)
    assert len(result) > 0
    # Verify all characters are in the base62 charset
    charset = '0123456789abcdefghijklmnopqrstuvwxyzABCEDFGHIJKLMNOPQRSTUVWXYZ'
    for char in result:
        assert char in charset


def test_load_private_key_invalid_type() -> None:
    """Test that load_private_key raises TypeError for non-RSA keys."""
    from unittest.mock import patch

    with (
        patch('apify._crypto.serialization.load_pem_private_key', return_value='not_an_rsa_key'),
        pytest.raises(TypeError, match='Invalid private key'),
    ):
        load_private_key('ZHVtbXk=', 'password')


def test_load_public_key_invalid_type() -> None:
    """Test that _load_public_key raises TypeError for non-RSA keys."""
    from unittest.mock import patch

    with (
        patch('apify._crypto.serialization.load_pem_public_key', return_value='not_an_rsa_key'),
        pytest.raises(TypeError, match='Invalid public key'),
    ):
        _load_public_key('ZHVtbXk=')


@pytest.mark.parametrize(
    'value',
    [
        pytest.param('just a string', id='string'),
        pytest.param(42, id='int'),
        pytest.param([1, 2, 3], id='list'),
        pytest.param(None, id='none'),
    ],
)
def test_decrypt_input_secrets_non_dict(value: object) -> None:
    """Test that decrypt_input_secrets returns non-dict input unchanged."""
    from apify._crypto import decrypt_input_secrets

    result = decrypt_input_secrets(PRIVATE_KEY, value)
    if value is None:
        assert result is None
    else:
        assert result == value



from __future__ import annotations

import asyncio

from apify import Actor, Request


async def custom_new_url_function(
    session_id: str | None = None,
    _: Request | None = None,
) -> str | None:
    if session_id is not None:
        return f'http://my-custom-proxy-supporting-sessions.com?session-id={session_id}'
    return 'http://my-custom-proxy-not-supporting-sessions.com'


async def main() -> None:
    async with Actor:
        proxy_cfg = await Actor.create_proxy_configuration(
            new_url_function=custom_new_url_function,  # ty: ignore[invalid-argument-type]
        )

        if not proxy_cfg:
            raise RuntimeError('No proxy configuration available.')

        proxy_url_with_session = await proxy_cfg.new_url('a')
        Actor.log.info(f'Using proxy URL: {proxy_url_with_session}')

        proxy_url_without_session = await proxy_cfg.new_url()
        Actor.log.info(f'Using proxy URL: {proxy_url_without_session}')


if __name__ == '__main__':
    asyncio.run(main())



from __future__ import annotations

from typing import TYPE_CHECKING

from .conftest import get_scrapy_source_files, read_actor_source, verify_spider_results

if TYPE_CHECKING:
    from ..conftest import MakeActorFunction, RunActorFunction

_EXPECTED_PRODUCTS = {
    'Widget A': {'price': '19.99', 'description': 'A basic widget for everyday use'},
    'Widget B': {'price': '29.99', 'description': 'An advanced widget with extra features'},
    'Widget C': {'price': '39.99', 'description': 'A premium widget for professionals'},
}


async def test_custom_pipeline_spider(make_actor: MakeActorFunction, run_actor: RunActorFunction) -> None:
    actor = await make_actor(
        label='scrapy-pipeline',
        source_files=get_scrapy_source_files(
            'spider_custom_pipeline.py',
            'CustomPipelineSpider',
            extra_source_files={
                'src/main.py': read_actor_source('main_custom_pipeline.py'),
                'src/settings_custom_pipeline.py': read_actor_source('settings_custom_pipeline.py'),
                'src/pipelines.py': read_actor_source('pipelines.py'),
            },
        ),
        additional_requirements=['scrapy>=2.14.0'],
    )
    run_result = await run_actor(actor)
    await verify_spider_results(actor, run_result, expected_products=_EXPECTED_PRODUCTS)



import asyncio

import httpx
from bs4 import BeautifulSoup

from apify import Actor


async def main() -> None:
    async with Actor:
        actor_input = await Actor.get_input() or {}
        url = actor_input.get('url', 'https://apify.com')
        async with httpx.AsyncClient() as client:
            response = await client.get(url)
        soup = BeautifulSoup(response.content, 'html.parser')
        data = {
            'url': url,
            'title': soup.title.string if soup.title else None,
        }
        await Actor.push_data(data)


if __name__ == '__main__':
    asyncio.run(main())



from __future__ import annotations

from crawlee._utils.crypto import crypto_random_object_id


def generate_unique_resource_name(label: str) -> str:
    """Generates a unique resource name, which will contain the given label."""
    name_template = 'python-sdk-tests-{}-generated-{}'
    template_length = len(name_template.format('', ''))
    api_name_limit = 63
    generated_random_id_length = 8
    label_length_limit = api_name_limit - template_length - generated_random_id_length

    label = label.replace('_', '-')
    assert len(label) <= label_length_limit, f'Max label length is {label_length_limit}, but got {len(label)}'

    return name_template.format(label, crypto_random_object_id(generated_random_id_length))



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from scrapy.linkextractors import LinkExtractor
from scrapy.spiders import CrawlSpider, Rule

if TYPE_CHECKING:
    from collections.abc import Generator

    from scrapy.http.response import Response


class CrawlProductSpider(CrawlSpider):
    name = 'crawl_product_spider'

    rules = (Rule(LinkExtractor(allow=r'/products/'), callback='parse_product'),)

    def __init__(self, start_urls: list[str], *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.start_urls = start_urls

    def parse_product(self, response: Response) -> Generator[dict, None, None]:
        yield {
            'url': response.url,
            'name': response.css('h1::text').get(''),
            'price': response.css('span.price::text').get(''),
            'description': response.css('p.description::text').get(''),
        }



from __future__ import annotations

import asyncio
import json
from pathlib import Path

import pytest

from crawlee import service_locator
from crawlee._consts import METADATA_FILENAME

from apify import Actor, Configuration
from apify.storage_clients._file_system import ApifyFileSystemKeyValueStoreClient


async def test_purge_preserves_input_file_and_metadata() -> None:
    """Test that purge() preserves INPUT.json and metadata files but removes other files."""
    # Get the global configuration (storage directory is set by test fixtures)
    configuration = Configuration.get_global_configuration()

    kvs_storage_client = await ApifyFileSystemKeyValueStoreClient.open(
        id=None,
        name='test-kvs',
        alias=None,
        configuration=configuration,
    )

    # Create some test files in the KVS directory
    kvs_path = kvs_storage_client.path_to_kvs

    # Create various files
    input_file = kvs_path / f'{configuration.input_key}'
    input_metadata_file = kvs_path / f'{configuration.input_key}.{METADATA_FILENAME}.json'
    metadata_file = kvs_path / METADATA_FILENAME
    regular_file1 = kvs_path / 'regular_file1.json'
    regular_file2 = kvs_path / 'another_file.txt'

    # Write content to files
    await asyncio.to_thread(input_file.write_text, '{"test": "input"}')
    await asyncio.to_thread(input_metadata_file.write_text, 'some text content')
    await asyncio.to_thread(regular_file1.write_text, '{"test": "data1"}')
    await asyncio.to_thread(regular_file2.write_text, 'some text content')

    # Verify all files exist before purge
    assert input_file.exists()
    assert input_metadata_file.exists()
    assert metadata_file.exists()  # Should exist from client creation
    assert regular_file1.exists()
    assert regular_file2.exists()

    # Purge the key-value store
    await kvs_storage_client.purge()

    # Verify INPUT.json and metadata are preserved
    assert input_file.exists(), f'{configuration.input_key} should be preserved during purge'
    assert metadata_file.exists(), f'{METADATA_FILENAME} should be preserved during purge'

    # Verify other files are deleted
    assert not regular_file1.exists(), 'Regular files should be deleted during purge'
    assert not regular_file2.exists(), 'Regular files should be deleted during purge'

    # Verify INPUT.json content is unchanged
    input_content = await asyncio.to_thread(input_file.read_text)
    assert json.loads(input_content) == json.loads('{"test": "input"}')


@pytest.mark.parametrize('input_file_name', ['INPUT', 'INPUT.json'])
async def test_pre_existing_input_used_by_actor(input_file_name: str) -> None:
    configuration = Configuration()
    service_locator.set_configuration(configuration)

    # Create key-value store directory and make sure that it is empty
    path_to_input = Path(configuration.storage_dir) / 'key_value_stores' / 'default'
    path_to_input.mkdir(parents=True)
    assert list(path_to_input.glob('*')) == []

    pre_existing_input = {
        'foo': 'bar',
    }

    # Create pre-existing INPUT.json file
    (path_to_input / input_file_name).write_text(json.dumps(pre_existing_input))

    async with Actor():
        assert pre_existing_input == await Actor.get_input()

    # Make sure that the input file doesn't get renamed in the process and metadata are added
    assert set(path_to_input.glob('*')) == {
        path_to_input / '__metadata__.json',
        path_to_input / input_file_name,
        path_to_input / f'{input_file_name}.__metadata__.json',
    }



"""Test HTTP server for e2e tests.

Serves an e-commerce test website with a category-based structure for testing crawl depth:

    / (depth 0) - Homepage with links to products, categories, about page, and deep chain
    /categories/electronics (depth 1) - Links to products 1 and 2
    /categories/home (depth 1) - Links to product 3
    /about (depth 1) - About page
    /deep/1 (depth 1) -> /deep/2 (depth 2) -> /deep/3 (depth 3) -> ... (infinite chain)
    /products/1 (depth 1 or 2) - Widget A
    /products/2 (depth 1 or 2) - Widget B
    /products/3 (depth 1 or 2) - Widget C

The homepage includes both direct product links (for Scrapy spiders that look for /products/ links
on the start page) and category links (for testing crawl depth with Crawlee crawlers).
With max_crawl_depth=2, the crawler reaches all products and categories but does not go beyond /deep/2.
"""

from __future__ import annotations

import asyncio
import logging
from collections.abc import Awaitable, Callable, Coroutine
from typing import Any

from uvicorn import Config
from uvicorn.server import Server

Receive = Callable[[], Awaitable[dict[str, Any]]]
Send = Callable[[dict[str, Any]], Coroutine[None, None, None]]

_PRODUCTS = {
    '1': {'name': 'Widget A', 'price': '$19.99', 'description': 'A basic widget for everyday use'},
    '2': {'name': 'Widget B', 'price': '$29.99', 'description': 'An advanced widget with extra features'},
    '3': {'name': 'Widget C', 'price': '$39.99', 'description': 'A premium widget for professionals'},
}


async def _send_html(send: Send, html: str, status: int = 200) -> None:
    await send(
        {
            'type': 'http.response.start',
            'status': status,
            'headers': [[b'content-type', b'text/html; charset=utf-8']],
        }
    )
    await send({'type': 'http.response.body', 'body': html.encode()})


async def app(scope: dict[str, Any], _receive: Receive, send: Send) -> None:
    assert scope['type'] == 'http'
    path = scope['path']

    if path == '/':
        await _send_html(
            send,
            '<html><head><title>E-commerce Test Store</title></head><body>'
            '<h1>Welcome to Test Store</h1>'
            '<a href="/products/1">Widget A</a>'
            '<a href="/products/2">Widget B</a>'
            '<a href="/products/3">Widget C</a>'
            '<a href="/categories/electronics">Electronics</a>'
            '<a href="/categories/home">Home &amp; Garden</a>'
            '<a href="/about">About Us</a>'
            '<a href="/deep/1">Explore More</a>'
            '</body></html>',
        )
    elif path == '/categories/electronics':
        await _send_html(
            send,
            '<html><head><title>Electronics</title></head><body>'
            '<h1>Electronics</h1>'
            '<a href="/products/1">Widget A</a>'
            '<a href="/products/2">Widget B</a>'
            '<a href="/">Back to Home</a>'
            '</body></html>',
        )
    elif path == '/categories/home':
        await _send_html(
            send,
            '<html><head><title>Home &amp; Garden</title></head><body>'
            '<h1>Home &amp; Garden</h1>'
            '<a href="/products/3">Widget C</a>'
            '<a href="/">Back to Home</a>'
            '</body></html>',
        )
    elif path.startswith('/products/'):
        product = _PRODUCTS.get(path.split('/')[-1])
        if product:
            await _send_html(
                send,
                f'<html><head><title>{product["name"]}</title></head><body>'
                f'<h1>{product["name"]}</h1>'
                f'<span class="price">{product["price"]}</span>'
                f'<p class="description">{product["description"]}</p>'
                f'<a href="/">Back to Home</a>'
                f'</body></html>',
            )
        else:
            await _send_html(send, '<html><body>Not Found</body></html>', 404)
    elif path == '/about':
        await _send_html(
            send,
            '<html><head><title>About Us</title></head><body>'
            '<h1>About Test Store</h1>'
            '<p class="description">We sell the best widgets in the world.</p>'
            '<a href="/">Back to Home</a>'
            '</body></html>',
        )
    elif path.startswith('/deep/'):
        try:
            n = int(path.split('/')[-1])
        except ValueError:
            await _send_html(send, '<html><body>Not Found</body></html>', 404)
            return
        await _send_html(
            send,
            f'<html><head><title>Deep Page {n}</title></head><body>'
            f'<h1>Deep Page {n}</h1>'
            f'<a href="/deep/{n + 1}">Go Deeper</a>'
            f'<a href="/">Back to Home</a>'
            f'</body></html>',
        )
    else:
        await _send_html(send, '<html><body>Not Found</body></html>', 404)


if __name__ == '__main__':
    asyncio.run(
        Server(
            config=Config(
                app=app,
                lifespan='off',
                loop='asyncio',
                port=8080,
                log_config=None,
                log_level=logging.CRITICAL,
            )
        ).serve()
    )



from apify import Actor


async def main() -> None:
    async with Actor:
        # Start the apify/screenshot-url Actor.
        actor_run = await Actor.call(
            actor_id='apify/screenshot-url',
            run_input={'url': 'http://example.com', 'delay': 10000},
        )

        if actor_run is None:
            raise RuntimeError('Actor task failed to start.')

        # Wait for the Actor run to finish.
        run_client = Actor.apify_client.run(actor_run.id)
        await run_client.wait_for_finish()

        # Get the Actor output from the key-value store.
        kvs_client = run_client.key_value_store()
        output = await kvs_client.get_record('OUTPUT')
        Actor.log.info(f'Actor output: {output}')



# ruff: noqa: ARG001 ARG005

from __future__ import annotations

import asyncio
import re
from dataclasses import asdict
from typing import TYPE_CHECKING, Any
from unittest.mock import Mock

import pytest

from apify_client import ApifyClientAsync
from apify_shared.consts import ApifyEnvVars

from apify._proxy_configuration import ProxyConfiguration, is_url

if TYPE_CHECKING:
    from pytest_httpserver import HTTPServer
    from werkzeug import Request, Response

    from .conftest import ApifyClientAsyncPatcher

DUMMY_PASSWORD = 'DUMMY_PASSWORD'


@pytest.fixture
def patched_apify_client(apify_client_async_patcher: ApifyClientAsyncPatcher) -> ApifyClientAsync:
    apify_client_async_patcher.patch(
        'user',
        'get',
        return_value={
            'proxy': {
                'password': DUMMY_PASSWORD,
            },
        },
    )
    return ApifyClientAsync()


def test_basic_constructor() -> None:
    groups = ['GROUP1', 'GROUP2']
    password = 'abcd1234'
    country_code = 'US'
    proxy_configuration = ProxyConfiguration(
        groups=groups,
        password=password,
        country_code=country_code,
    )
    assert proxy_configuration._groups == groups
    assert proxy_configuration._password == password
    assert proxy_configuration._country_code == country_code


def test_fallback_constructor(monkeypatch: pytest.MonkeyPatch) -> None:
    hostname = 'example.com'
    password = 'abcd1234'
    port = 1234

    monkeypatch.setenv('APIFY_PROXY_HOSTNAME', hostname)
    monkeypatch.setenv('APIFY_PROXY_PASSWORD', password)
    monkeypatch.setenv('APIFY_PROXY_PORT', f'{port}')

    proxy_configuration = ProxyConfiguration()

    assert proxy_configuration._hostname == hostname
    assert proxy_configuration._password == password
    assert proxy_configuration._port == port


def test_invalid_arguments() -> None:
    for invalid_groups, bad_group_index in [
        (['abc', 'de-f', 'geh'], 1),
        (['', 'def', 'geh'], 0),
        (['abc', 'DEF', 'geh$'], 2),
        ([111, 'DEF', 'geh$'], 2),
    ]:
        bad_group = str(invalid_groups[bad_group_index])

        # Match the actual error message pattern that includes the value and argument name
        match_pattern = f'Value {re.escape(bad_group)} of argument groups does not match pattern'

        with pytest.raises(ValueError, match=match_pattern):
            ProxyConfiguration(groups=invalid_groups)  # ty: ignore[invalid-argument-type]

    for invalid_country_code in ['CZE', 'aa', 'DDDD', 1111]:
        match_pattern = f'Value {re.escape(str(invalid_country_code))} of argument country_code does not match pattern'
        with pytest.raises(ValueError, match=match_pattern):
            ProxyConfiguration(country_code=invalid_country_code)  # ty: ignore[invalid-argument-type]

    with pytest.raises(ValueError, match=r'Exactly one of .* must be specified'):
        ProxyConfiguration(
            proxy_urls=['http://proxy.com:1111'],
            new_url_function=lambda session_id=None, request=None: 'http://proxy.com:2222',
        )

    with pytest.raises(ValueError, match=r'Cannot combine custom proxies with Apify Proxy'):
        ProxyConfiguration(proxy_urls=['http://proxy.com:1111'], groups=['GROUP1'])

    with pytest.raises(ValueError, match=re.escape('bad-url')):
        ProxyConfiguration(proxy_urls=['bad-url'])

    with pytest.raises(ValueError, match=r'Cannot combine custom proxies with Apify Proxy'):
        ProxyConfiguration(
            new_url_function=lambda session_id=None, request=None: 'http://proxy.com:2222', groups=['GROUP1']
        )


async def test_new_url_basic() -> None:
    groups = ['GROUP1', 'GROUP2']
    password = 'abcd1234'
    country_code = 'US'
    proxy_configuration = ProxyConfiguration(
        groups=groups,
        password=password,
        country_code=country_code,
    )
    proxy_url = await proxy_configuration.new_url()

    expected_username = f'groups-{"+".join(groups)},country-{country_code}'
    expected_hostname = 'proxy.apify.com'
    expected_port = 8000

    assert proxy_url == f'http://{expected_username}:{password}@{expected_hostname}:{expected_port}'


async def test_new_url_with_session_ids() -> None:
    groups = ['GROUP1', 'GROUP2']
    password = 'abcd1234'
    country_code = 'US'
    proxy_configuration = ProxyConfiguration(
        groups=groups,
        password=password,
        country_code=country_code,
    )

    session_ids: list[str] = [
        'a',
        'a_b',
        'a_2',
        'a_1_b',
        'aaa~BBB',
        '1',
        '0.34252352',
        '123456',
        'XXXXXXXXXXxxxxxxxxxxXXXXXXXXXXxxxxxxxxxxXXXXXXXXXX',
    ]
    for session_id in session_ids:
        expected_username = f'groups-{"+".join(groups)},session-{session_id},country-{country_code}'
        expected_hostname = 'proxy.apify.com'
        expected_port = 8000

        proxy_url = await proxy_configuration.new_url(session_id)

        assert proxy_url == f'http://{expected_username}:{password}@{expected_hostname}:{expected_port}'

    for invalid_session_id in ['a-b', 'a$b', 'XXXXXXXXXXxxxxxxxxxxXXXXXXXXXXxxxxxxxxxxXXXXXXXXXXTooLong']:
        with pytest.raises(ValueError, match=re.escape(str(invalid_session_id))):
            await proxy_configuration.new_url(invalid_session_id)


async def test_rotating_custom_urls() -> None:
    proxy_urls: list[str | None] = ['http://proxy.com:1111', 'http://proxy.com:2222', 'http://proxy.com:3333']
    proxy_configuration = ProxyConfiguration(proxy_urls=proxy_urls)

    assert await proxy_configuration.new_url() == proxy_urls[0]
    assert await proxy_configuration.new_url() == proxy_urls[1]
    assert await proxy_configuration.new_url() == proxy_urls[2]
    assert await proxy_configuration.new_url() == proxy_urls[0]
    assert await proxy_configuration.new_url() == proxy_urls[1]
    assert await proxy_configuration.new_url() == proxy_urls[2]


async def test_rotating_custom_urls_with_sessions() -> None:
    sessions = ['session_01', 'session_02', 'session_03', 'session_04', 'session_05', 'session_06']
    proxy_urls: list[str | None] = ['http://proxy.com:1111', 'http://proxy.com:2222', 'http://proxy.com:3333']

    proxy_configuration = ProxyConfiguration(proxy_urls=proxy_urls)

    # same session should use same proxy URL
    assert await proxy_configuration.new_url(sessions[0]) == proxy_urls[0]
    assert await proxy_configuration.new_url(sessions[0]) == proxy_urls[0]
    assert await proxy_configuration.new_url(sessions[0]) == proxy_urls[0]

    # different sessions should rotate different proxies
    assert await proxy_configuration.new_url(sessions[1]) == proxy_urls[1]
    assert await proxy_configuration.new_url(sessions[2]) == proxy_urls[2]
    assert await proxy_configuration.new_url(sessions[3]) == proxy_urls[0]
    assert await proxy_configuration.new_url(sessions[4]) == proxy_urls[1]
    assert await proxy_configuration.new_url(sessions[5]) == proxy_urls[2]

    # already used sessions should be remembered
    assert await proxy_configuration.new_url(sessions[1]) == proxy_urls[1]
    assert await proxy_configuration.new_url(sessions[3]) == proxy_urls[0]


async def test_with_custom_new_url_function() -> None:
    custom_urls = [
        'http://proxy.com:1111',
        'http://proxy.com:2222',
        'http://proxy.com:3333',
        'http://proxy.com:4444',
        'http://proxy.com:5555',
        'http://proxy.com:6666',
    ]

    def custom_new_url_function(session_id: str | None = None, request: Any = None) -> str:
        nonlocal custom_urls
        return custom_urls.pop()

    proxy_configuration = ProxyConfiguration(new_url_function=custom_new_url_function)

    for custom_url in reversed(custom_urls):
        assert await proxy_configuration.new_url() == custom_url


async def test_with_async_custom_new_url_function() -> None:
    custom_urls = [
        'http://proxy.com:1111',
        'http://proxy.com:2222',
        'http://proxy.com:3333',
        'http://proxy.com:4444',
        'http://proxy.com:5555',
        'http://proxy.com:6666',
    ]

    async def custom_new_url_function(session_id: str | None = None, request: Any = None) -> str:
        nonlocal custom_urls
        await asyncio.sleep(0.1)
        return custom_urls.pop()

    proxy_configuration = ProxyConfiguration(new_url_function=custom_new_url_function)

    for custom_url in reversed(custom_urls):
        assert await proxy_configuration.new_url() == custom_url


async def test_invalid_custom_new_url_function() -> None:
    def custom_new_url_function(session_id: str | None = None, request: Any = None) -> str:
        raise ValueError

    proxy_configuration = ProxyConfiguration(new_url_function=custom_new_url_function)

    with pytest.raises(ValueError, match=r'The provided "new_url_function" did not return a valid URL'):
        await proxy_configuration.new_url()


async def test_url_reference_not_shared_between_instances() -> None:
    proxy_urls: list[str | None] = [
        'http://proxy-example-1.com:8000',
        'http://proxy-example-2.com:8000',
    ]
    proxy_configuration_1 = ProxyConfiguration(proxy_urls=proxy_urls)

    proxy_urls.append('http://proxy-example-3.com:8000')
    proxy_configuration_2 = ProxyConfiguration(proxy_urls=proxy_urls)

    assert proxy_configuration_1._proxy_urls is not proxy_configuration_2._proxy_urls

    session_id = 'ABCD'
    await proxy_configuration_1.new_url(session_id=session_id)
    await proxy_configuration_2.new_url(session_id=session_id)

    assert proxy_configuration_1._used_proxy_urls is not proxy_configuration_2._used_proxy_urls


async def test_new_proxy_info_basic_construction() -> None:
    groups = ['GROUP1', 'GROUP2']
    password = 'abcd1234'
    country_code = 'US'
    proxy_configuration = ProxyConfiguration(
        groups=groups,
        password=password,
        country_code=country_code,
    )

    proxy_info = await proxy_configuration.new_proxy_info()
    assert proxy_info is not None

    expected_hostname = 'proxy.apify.com'
    expected_port = 8000
    expected_username = f'groups-{"+".join(groups)},country-{country_code}'

    assert asdict(proxy_info) == {
        'url': f'http://{expected_username}:{password}@{expected_hostname}:{expected_port}',
        'hostname': expected_hostname,
        'port': expected_port,
        'groups': groups,
        'country_code': country_code,
        'username': expected_username,
        'password': password,
        'proxy_tier': None,
        'session_id': None,
        'scheme': 'http',
    }


async def test_new_proxy_info_rotating_urls() -> None:
    proxy_urls: list[str | None] = ['http://proxy.com:1111', 'http://proxy.com:2222', 'http://proxy.com:3333']
    proxy_configuration = ProxyConfiguration(proxy_urls=proxy_urls)

    proxy_info = await proxy_configuration.new_proxy_info()
    assert proxy_info is not None
    assert proxy_info.url == proxy_urls[0]

    proxy_info = await proxy_configuration.new_proxy_info()
    assert proxy_info is not None
    assert proxy_info.url == proxy_urls[1]

    proxy_info = await proxy_configuration.new_proxy_info()
    assert proxy_info is not None
    assert proxy_info.url == proxy_urls[2]

    proxy_info = await proxy_configuration.new_proxy_info()
    assert proxy_info is not None
    assert proxy_info.url == proxy_urls[0]

    proxy_info = await proxy_configuration.new_proxy_info()
    assert proxy_info is not None
    assert proxy_info.url == proxy_urls[1]

    proxy_info = await proxy_configuration.new_proxy_info()
    assert proxy_info is not None
    assert proxy_info.url == proxy_urls[2]


async def test_new_proxy_info_rotating_urls_with_sessions() -> None:
    sessions = ['session_01', 'session_02', 'session_03', 'session_04', 'session_05', 'session_06']
    proxy_urls: list[str | None] = ['http://proxy.com:1111', 'http://proxy.com:2222', 'http://proxy.com:3333']

    proxy_configuration = ProxyConfiguration(proxy_urls=proxy_urls)

    # same session should use same proxy URL
    proxy_info = await proxy_configuration.new_proxy_info(sessions[0])
    assert proxy_info is not None
    assert proxy_info.url == proxy_urls[0]

    proxy_info = await proxy_configuration.new_proxy_info(sessions[0])
    assert proxy_info is not None
    assert proxy_info.url == proxy_urls[0]

    proxy_info = await proxy_configuration.new_proxy_info(sessions[0])
    assert proxy_info is not None
    assert proxy_info.url == proxy_urls[0]

    # different sessions should rotate different proxies
    proxy_info = await proxy_configuration.new_proxy_info(sessions[1])
    assert proxy_info is not None
    assert proxy_info.url == proxy_urls[1]

    proxy_info = await proxy_configuration.new_proxy_info(sessions[2])
    assert proxy_info is not None
    assert proxy_info.url == proxy_urls[2]

    proxy_info = await proxy_configuration.new_proxy_info(sessions[3])
    assert proxy_info is not None
    assert proxy_info.url == proxy_urls[0]

    proxy_info = await proxy_configuration.new_proxy_info(sessions[4])
    assert proxy_info is not None
    assert proxy_info.url == proxy_urls[1]

    proxy_info = await proxy_configuration.new_proxy_info(sessions[5])
    assert proxy_info is not None
    assert proxy_info.url == proxy_urls[2]

    # already used sessions should be remembered
    proxy_info = await proxy_configuration.new_proxy_info(sessions[1])
    assert proxy_info is not None
    assert proxy_info.url == proxy_urls[1]

    proxy_info = await proxy_configuration.new_proxy_info(sessions[3])
    assert proxy_info is not None
    assert proxy_info.url == proxy_urls[0]


@pytest.mark.usefixtures('patched_impit_client')
async def test_initialize_with_valid_configuration(
    monkeypatch: pytest.MonkeyPatch,
    httpserver: HTTPServer,
    patched_apify_client: ApifyClientAsync,
) -> None:
    dummy_proxy_status_url = str(httpserver.url_for('/')).removesuffix('/')
    monkeypatch.setenv(ApifyEnvVars.TOKEN.value, 'DUMMY_TOKEN')
    monkeypatch.setenv(ApifyEnvVars.PROXY_STATUS_URL.value, dummy_proxy_status_url)

    call_mock = Mock()

    def request_handler(request: Request, response: Response) -> Response:
        call_mock(request.url)
        return response

    httpserver.expect_oneshot_request('/').with_post_hook(request_handler).respond_with_json(
        {
            'connected': True,
            'connectionError': None,
            'isManInTheMiddle': True,
        },
        status=200,
    )

    proxy_configuration = ProxyConfiguration(_apify_client=patched_apify_client)

    await proxy_configuration.initialize()

    assert proxy_configuration._password == DUMMY_PASSWORD
    assert proxy_configuration.is_man_in_the_middle is True

    assert len(patched_apify_client.calls['user']['get']) == 1  # ty:ignore[unresolved-attribute]
    assert call_mock.call_count == 1


async def test_initialize_without_password_or_token() -> None:
    proxy_configuration = ProxyConfiguration()

    with pytest.raises(ValueError, match=r'Apify Proxy password must be provided'):
        await proxy_configuration.initialize()


@pytest.mark.usefixtures('patched_impit_client')
async def test_initialize_with_manual_password(monkeypatch: pytest.MonkeyPatch, httpserver: HTTPServer) -> None:
    dummy_proxy_status_url = str(httpserver.url_for('/')).removesuffix('/')
    monkeypatch.setenv(ApifyEnvVars.PROXY_STATUS_URL.value, dummy_proxy_status_url)

    httpserver.expect_oneshot_request('/').respond_with_json(
        {
            'connected': True,
            'connectionError': None,
            'isManInTheMiddle': False,
        },
        status=200,
    )

    proxy_configuration = ProxyConfiguration(password=DUMMY_PASSWORD)

    await proxy_configuration.initialize()

    assert proxy_configuration._password == DUMMY_PASSWORD
    assert proxy_configuration.is_man_in_the_middle is False


@pytest.mark.usefixtures('patched_impit_client')
async def test_initialize_preferring_password_from_env_over_calling_api(
    monkeypatch: pytest.MonkeyPatch,
    httpserver: HTTPServer,
    patched_apify_client: ApifyClientAsync,
) -> None:
    dummy_proxy_status_url = str(httpserver.url_for('/')).removesuffix('/')
    monkeypatch.setenv(ApifyEnvVars.PROXY_STATUS_URL.value, dummy_proxy_status_url)
    monkeypatch.setenv(ApifyEnvVars.PROXY_PASSWORD.value, DUMMY_PASSWORD)

    httpserver.expect_oneshot_request('/').respond_with_json(
        {
            'connected': True,
            'connectionError': None,
            'isManInTheMiddle': False,
        },
        status=200,
    )

    proxy_configuration = ProxyConfiguration()

    await proxy_configuration.initialize()

    assert proxy_configuration._password == DUMMY_PASSWORD
    assert proxy_configuration.is_man_in_the_middle is False

    assert len(patched_apify_client.calls['user']['get']) == 0  # ty: ignore[unresolved-attribute]


@pytest.mark.usefixtures('patched_impit_client')
async def test_initialize_with_manual_password_different_than_user_one(
    monkeypatch: pytest.MonkeyPatch,
    httpserver: HTTPServer,
    patched_apify_client: ApifyClientAsync,
) -> None:
    dummy_proxy_status_url = str(httpserver.url_for('/')).removesuffix('/')
    different_dummy_password = 'DIFFERENT_DUMMY_PASSWORD'
    monkeypatch.setenv(ApifyEnvVars.TOKEN.value, 'DUMMY_TOKEN')
    monkeypatch.setenv(ApifyEnvVars.PROXY_STATUS_URL.value, dummy_proxy_status_url)
    monkeypatch.setenv(ApifyEnvVars.PROXY_PASSWORD.value, different_dummy_password)

    httpserver.expect_oneshot_request('/').respond_with_json(
        {
            'connected': True,
            'connectionError': None,
            'isManInTheMiddle': True,
        },
        status=200,
    )

    proxy_configuration = ProxyConfiguration(_apify_client=patched_apify_client)

    await proxy_configuration.initialize()

    assert proxy_configuration._password == different_dummy_password
    assert proxy_configuration.is_man_in_the_middle is True


@pytest.mark.usefixtures('patched_impit_client')
async def test_initialize_when_not_connected(monkeypatch: pytest.MonkeyPatch, httpserver: HTTPServer) -> None:
    dummy_connection_error = 'DUMMY_CONNECTION_ERROR'
    dummy_proxy_status_url = str(httpserver.url_for('/')).removesuffix('/')
    monkeypatch.setenv(ApifyEnvVars.PROXY_STATUS_URL.value, dummy_proxy_status_url)

    httpserver.expect_oneshot_request('/').respond_with_json(
        {
            'connected': False,
            'connectionError': dummy_connection_error,
        },
        status=200,
    )

    proxy_configuration = ProxyConfiguration(password=DUMMY_PASSWORD)

    with pytest.raises(ConnectionError, match=dummy_connection_error):
        await proxy_configuration.initialize()


async def test_initialize_when_status_page_unavailable(
    monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture, httpserver: HTTPServer
) -> None:
    caplog.set_level('WARNING')
    dummy_proxy_status_url = str(httpserver.url_for('/')).removesuffix('/')
    monkeypatch.setenv(ApifyEnvVars.PROXY_STATUS_URL.value, dummy_proxy_status_url)

    httpserver.expect_oneshot_request('/').respond_with_data(status=500)

    proxy_configuration = ProxyConfiguration(password=DUMMY_PASSWORD)

    await proxy_configuration.initialize()

    assert (
        'Apify Proxy access check timed out. Watch out for errors with status code 407. If you see some, it most likely'
        ' means you do not have access to either all or some of the proxies you are trying to use.'
    ) in caplog.messages


async def test_initialize_with_non_apify_proxy(
    monkeypatch: pytest.MonkeyPatch,
    httpserver: HTTPServer,
    patched_apify_client: ApifyClientAsync,
) -> None:
    dummy_proxy_status_url = str(httpserver.url_for('/')).removesuffix('/')
    monkeypatch.setenv(ApifyEnvVars.PROXY_STATUS_URL.value, dummy_proxy_status_url)

    call_mock = Mock()

    def request_handler(request: Request, response: Response) -> Response:
        call_mock(request.url)
        return response

    httpserver.expect_oneshot_request('/').with_post_hook(request_handler).respond_with_data(status=200)

    proxy_configuration = ProxyConfiguration(proxy_urls=['http://dummy-proxy.com:8000'])

    await proxy_configuration.initialize()

    assert len(patched_apify_client.calls['user']['get']) == 0  # ty: ignore[unresolved-attribute]
    assert call_mock.call_count == 0


def test_is_url_validation() -> None:
    assert is_url('http://dummy-proxy.com:8000') is True
    assert is_url('https://example.com') is True
    assert is_url('http://localhost') is True
    assert is_url('https://12.34.56.78') is True
    assert is_url('http://12.34.56.78:9012') is True
    assert is_url('http://::1') is True
    assert is_url('https://2f45:4da6:8f56:af8c:5dce:c1de:14d2:8661') is True

    assert is_url('dummy-proxy.com:8000') is False
    assert is_url('gyfwgfhkjhljkfhdsf') is False
    assert is_url('http://') is False
    assert is_url('http://example') is False
    assert is_url('http:/example.com') is False
    assert is_url('12.34.56.78') is False
    assert is_url('::1') is False
    assert is_url('https://4da6:8f56:af8c:5dce:c1de:14d2:8661') is False


@pytest.mark.parametrize(
    'value',
    [
        pytest.param('', id='empty_string'),
        pytest.param(None, id='none'),
    ],
)
def test_is_url_with_completely_unparsable_input(value: str | None) -> None:
    """Test is_url with input that causes urlparse to fail."""
    assert is_url(value) is False  # ty: ignore[invalid-argument-type]


def test_check_min_length_raises() -> None:
    """Test that _check raises ValueError for values shorter than min_length."""
    from apify._proxy_configuration import _check

    with pytest.raises(ValueError, match='shorter than minimum allowed length'):
        _check('ab', label='session_id', min_length=5)


async def test_proxy_urls_with_apify_domain_warns(caplog: pytest.LogCaptureFixture) -> None:
    """Test that using apify.com URLs in proxy_urls triggers a warning."""
    caplog.set_level('WARNING')
    ProxyConfiguration(proxy_urls=['http://proxy.apify.com:8000'])
    assert any('Some Apify proxy features may work incorrectly' in msg for msg in caplog.messages)
